#!/usr/bin/python3

from sys import argv
from novowrap.assembly import assembly_main
from novowrap.utils import get_all_third_party
import logging


# define logger
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
log = logging.getLogger('novowrap')
try:
    import coloredlogs
    coloredlogs.install(level=logging.INFO, fmt=FMT, datefmt=DATEFMT)
except ImportError:
    pass
log = logging.getLogger('novowrap')


def main():
    if argv[-1] == 'init':
        get_all_third_party()
        return
    elif argv[-1] == '-h':
        assembly_main()
        return
    elif len(argv) > 1:
        assembly_main()
        return
    try:
        from novowrap.ui import ui_main
        ui_main()
    except Exception:
        assembly_main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

from collections import defaultdict, deque
from itertools import product as cartesian_product
from pathlib import Path
from random import shuffle
import argparse
import logging

from Bio import SeqIO
from graphviz import Digraph

from novowrap.utils import blast, parse_blast_tab, test_cmd

PREFIX = '_RC_'

# define logger
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
try:
    import coloredlogs
    coloredlogs.install(level=logging.INFO, fmt=FMT, datefmt=DATEFMT)
except ImportError:
    pass
# inherit logger from novowrap
log = logging.getLogger('novowrap')
# test dot
if test_cmd('dot', '-V'):
    have_dot = True
else:
    have_dot = False
    log.debug('Cannot find dot, skip graphviz.')


def parse_args(arg_list=None):
    arg = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    arg.add_argument('-input', nargs='+', required=True,
                     help='input filenames')
    arg.add_argument('-out', help='output filename')
    if arg_list is None:
        return arg.parse_args()
    else:
        return arg.parse_args(arg_list)


def reverse_link(link):
    """
    Reverse blast_result
    """
    new = []
    for i in reversed(link):
        if i.startswith(PREFIX):
            i = i.replace(PREFIX, '')
        else:
            i = PREFIX + i
        new.append(i)
    return tuple(new)


def get_path(overlap):
    """
    Get path from overlap information.
    Args:
        overlap(list(blast_result)): overlaps
    Yield:
        path(list(blast_result)): ordered path
        is_circle(bool): is circle or not (linear)
    """
    scaffold = deque()
    overlap_dict = {i[0]: i for i in overlap}
    up_dict = {i[0]: i for i in overlap}
    down_dict = {i[1]: i for i in up_dict.values()}
    try:
        scaffold.append(overlap_dict.popitem()[1])
    except KeyError:
        raise Exception
    while True:
        try:
            up_name = scaffold[0][0]
            down_name = scaffold[-1][1]
        # name is None
        except TypeError:
            up_name = None
            down_name = None
        # if downstream not in overlap_dict:
        if up_name is None:
            pass
        elif up_name in down_dict:
            upstream = down_dict[up_name][0]
            if upstream in overlap_dict:
                scaffold.appendleft(overlap_dict.pop(upstream))
            else:
                scaffold.appendleft([None, None])
        else:
            # [None, None] as head/tail
            scaffold.appendleft([None, None])
        if down_name is None:
            pass
        elif down_name in up_dict:
            downstream = up_dict[down_name][0]
            if downstream in overlap_dict:
                scaffold.append(overlap_dict.pop(downstream))
            else:
                scaffold.append([None, None])
        else:
            scaffold.append([None, None])
        if scaffold[0][0] is None and scaffold[-1][0] is None:
            scaffold = list(scaffold)[1:-1]
            is_circle = (scaffold[0][0] == scaffold[-1][1])
            yield scaffold, is_circle
            scaffold = deque()
            try:
                scaffold.append(overlap_dict.popitem()[1])
            except KeyError:
                break


def get_contig(files, out):
    """
    Get contigs (SeqRecord) from input files.
    Add reverse-complement of each contigs.
    Args:
        files(list(str)): input files
        out(Path): output file name
    Return:
        contigs_and_rc(list(SeqRecord)): contigs with their reverse-complements
        contigs_and_rc_fasta(Path): fasta file
    """
    contigs_and_rc = []
    for f in files:
        fasta = Path(f).absolute()
        # some id may already used PREFIX
        fasta_stem = fasta.stem.replace('.fasta', '').replace(PREFIX, '-RC-')
        for idx, record in enumerate(SeqIO.parse(fasta, 'fasta')):
            record.id = f'{fasta_stem}-{idx}'
            record.description = ''
            contigs_and_rc.append(record)
            record_rc = record.reverse_complement(id=PREFIX+record.id)
            contigs_and_rc.append(record_rc)
    # print(necessary?)
    shuffle(contigs_and_rc)
    log.info(f'Got {len(contigs_and_rc)//2} contigs from input files.')
    contigs_and_rc_fasta = out.with_suffix('.with_rc.fasta')
    SeqIO.write(contigs_and_rc, contigs_and_rc_fasta, 'fasta')
    return contigs_and_rc, contigs_and_rc_fasta


def get_overlap(contigs, contigs_and_rc_fasta):
    """
    Get overlap by BLAST.
    Args:
        contigs_and_rc_fasta(Path): fasta file
        contigs(list(SeqRecord)): contigs
    Return:
        link_info(list(blast_result)): link info of contigs
    """
    blast_result, blast_log = blast(contigs_and_rc_fasta, contigs_and_rc_fasta)
    if blast_result is None:
        return []
    overlap = []
    for query, sequence in zip(parse_blast_tab(blast_result), contigs):
        if len(query) == 0:
            continue
        ambiguous_base_n = len(sequence.seq.strip('ATCGatcg'))
        p_ident_min = int((1-(ambiguous_base_n/len(sequence)))*100)
        for hit in query:
            (qseqid, sseqid, sstrand, qlen, slen, length, pident, gapopen,
             qstart, qend, sstart, send) = hit
            if qseqid == sseqid:
                continue
            if pident < p_ident_min:
                continue
            # only allow overlapped seq
            # nested seqs should be omit
            if qlen == length or slen == length:
                continue
            # normally, subject must be object's downstream
            # for tail->head, may be partial match
            # ambiguous base should not at head/tail
            if sstrand == 'plus':
                if sstart == 1 and qend == qlen:
                    pass
                else:
                    continue
            else:
                # skip minus
                continue
            overlap.append(hit)
    blast_result.unlink()
    blast_log.unlink()
    return overlap


def clean_overlap(overlap):
    """
    Remove transitively-inferrible edges ("shortcuts").
    Remove edges across two circle ("shortcuts_b").
    Remove non-branching stretches ("tips").
    Remove alternative path ("bubble").
    Arg:
        overlap(list(blast_result)): link info of contigs
    Return:
        cleaned_overlap(list(blast_result)): clean link
        edges(dict(set(dot.edge))): edges of overlap
    """
    def get_dict(overlap_list):
        up_down = defaultdict(set)
        down_up = defaultdict(set)
        for i in overlap_list:
            up_down[i[0]].add(i[1])
            down_up[i[1]].add(i[0])
        return up_down, down_up

    raw = {(i[0], i[1]): i for i in overlap}
    up_down, down_up = get_dict(overlap)
    # shortcuts between two circles
    between = set()
    for down, up in down_up.items():
        if len(up) == 1:
            continue
        for u in up:
            between.add((u, down))
    shortcuts_b = set()
    for i in between:
        if reverse_link(i) in between:
            shortcuts_b.add(i)
            shortcuts_b.add(reverse_link(i))
    # shortcuts_b in another direction
    between2 = set()
    for up, down in up_down.items():
        if len(down) == 1:
            continue
        for d in down:
            between.add((up, d))
    for i in between2:
        if reverse_link(i) in between2:
            shortcuts_b.add(i)
            shortcuts_b.add(reverse_link(i))
    # Remove transitively-inferible edges that across one contig
    # One step is enough, if longer, may be alternative path
    shortcuts = set()
    for up, down in up_down.items():
        if len(down) == 1:
            continue
        down_down = set()
        for d in down:
            if d in up_down:
                down_down.update(up_down[d])
        shortcuts.update({(up, i) for i in down_down & down})
    exclude = shortcuts & shortcuts_b
    shortcuts = shortcuts - exclude
    shortcuts_b = shortcuts_b - exclude
    # to_remove = shortcuts | short_tips | shortcuts_b
    to_remove = shortcuts | shortcuts_b
    cleaned_overlap = [raw[i] for i in raw if i not in to_remove]
    # a-b-c, a-d-c
    # or a-b-c-d, a-e
    bubble_path = []
    # long tip, a-b-c-d-a, b-d
    tips = set()
    # other types may ganrao bubble detection
    up_down, down_up = get_dict(cleaned_overlap)
    depth = len(up_down) - 1
    for up, down in up_down.items():
        if len(down) <= 1:
            continue
        path = []
        for d in down:
            step = 0
            p = [up, d]
            while step <= depth:
                # more than one upstream
                if len(down_up[d]) > 1:
                    break
                # tail of linear
                if d not in up_down:
                    break
                if up == d:
                    good = tuple(p[:2])
                    bad = {(up, i) for i in up_down[d]}
                    bad.remove(good)
                    r = {reverse_link(i) for i in bad}
                    tips.update(bad)
                    tips.update(r)
                    break
                if len(up_down[d]) > 1:
                    if up in up_down[d]:
                        t = set(up_down[d])
                        t.remove(up)
                        if t is not None and len(t) != 0:
                            tips.update(set((d, dd) for dd in t))
                            r = set((reverse_link((d, dd)) for dd in t))
                            tips.update(r)
                    break
                d = set(up_down[d])
                d = d.pop()
                p.append(d)
                step += 1
            path.append(p)
        n_tail = len(set([p[-1] for p in path]))
        n_length = len(set([len(p) for p in path]))
        if len(path) == 0:
            continue
        # same length, same head/tail
        if n_tail == 1 and n_length == 1:
            # if bubble, keep the first
            bubble_path.extend(path[1:])
    # tips in another direction
    for down, up in down_up.items():
        if len(up) <= 1:
            continue
        for u in up:
            step = 0
            p = [down, u]
            while step <= depth:
                # more than one downstream
                if len(up_down[u]) > 1:
                    break
                # head of linear
                if u not in down_up:
                    break
                if down == u:
                    good = (p[1], p[0])
                    bad = {(i, down) for i in down_up[u]}
                    bad.remove(good)
                    r = {reverse_link(i) for i in bad}
                    tips.update(bad)
                    tips.update(r)
                u = set(down_up[u])
                u = u.pop()
                p.append(u)
                step += 1
    bubble = set()
    for path in bubble_path:
        for i in range(len(path)-1):
            bubble.add((path[i], path[i+1]))
    to_remove = to_remove.union(bubble).union(tips)
    to_remove = to_remove.union(bubble)
    cleaned_overlap = [raw[i] for i in raw if i not in to_remove]
    edges = {'shortcuts': shortcuts, 'tips': tips,
             'shortcuts_b': shortcuts_b, 'bubble': bubble,
             'exclude': exclude}
    return cleaned_overlap, edges


def get_link(contigs_and_rc, contigs_and_rc_fasta):
    """
    Get overlap between contigs.
    Remove overlap that may cause chaos.
    Args:
        contigs_and_rc(list(SeqRecord)): contigs with their reverse-complements
        contigs_and_rc_fasta(Path): fasta file
    Return:
        link(list(blast_result)): link info of contigs
    """
    MAX_TRY = 2 ** 16
    overlap = get_overlap(contigs_and_rc, contigs_and_rc_fasta)
    if len(overlap) == 0:
        log.critical('Bad input. Failed to merge.')
        return []
    overlap_clean, edges = clean_overlap(overlap)
    if len(overlap_clean) == 0:
        log.critical('Bad input. Failed to merge.')
        return []
    overlap_clean_dict = {(i[0], i[1]): i for i in overlap_clean}
    for i in edges:
        log.debug(f'{i}, {len(edges[i])}')
    up_dict = defaultdict(set)
    down_dict = defaultdict(set)
    for i in overlap_clean:
        up_dict[i[0]].add(i[1])
        down_dict[i[1]].add(i[0])
    links = []
    possible = []
    up_down_clean = []
    for up, down in up_dict.items():
        if len(down) > 1:
            possible.append([overlap_clean_dict[(up, d)] for d in down])
        else:
            up_down_clean.append(overlap_clean_dict[(up, down.pop())])
    n = 0
    if len(possible) > MAX_TRY:
        log.debug(f'Too many possible ({len(possible)}).')
    for i in cartesian_product(*possible):
        if n >= MAX_TRY:
            break
        n += 1
        combine = up_down_clean + list(i)
        for link in get_path(combine):
            links.append(link)
    edges['link'] = set()
    for i in links:
        for j in i[0]:
            edges['link'].add((j[0], j[1]))
    # draw
    dot_out = contigs_and_rc_fasta.with_suffix('')
    if have_dot:
        dot = Digraph(engine='dot', node_attr={'shape': 'cds'})
        for i in overlap:
            dot.node(i[0])
            dot.node(i[1])
        # dot.edge(i[0], i[1], color='#999999')
        for edge in edges['link']:
            dot.edge(*edge, color='blue')
        for edge in edges['shortcuts']:
            dot.edge(*edge, color='red')
        for edge in edges['tips']:
            dot.edge(*edge, color='green')
        for edge in edges['shortcuts_b']:
            dot.edge(*edge, color='orange')
        for edge in edges['bubble']:
            dot.edge(*edge, color='purple')
        dot.render(dot_out)
        dot_out.unlink()
    else:
        log.debug('Cannot find graphviz, skip drawing figure.')
    # for i in links: print('->'.join([str((j[0], j[1])) for j in i[0]]))
    return links


def merge_seq(contigs, links):
    """
    Use overlap information of contigs to merge them.
    Assume link_info only contains one kind of link route.
    Arg:
        contigs(list(SeqRecord)): contigs
        links(list(blast_result, is_circle)): link info of contigs
    Return:
        merged(list(SeqRecord)): list of merged sequences
    """
    # give 2 circular result at most
    max_circle = 2
    # break if too much linear
    max_linear = 20
    contigs_d = {i.id: i for i in contigs}
    circular = []
    linear_d = {}
    for link, is_circle in links:
        if len(circular) >= max_circle:
            break
        elif len(linear_d) >= max_linear:
            log.debug(f'Cannot find circular assembly. Instead found '
                      '{len(circular)} linear asseblies. Break.')
            break
        # qseqid, sseqid, sstrand, qlen, slen, length, pident, gapopen, qstart,
        # qend, sstart, send
        seq = contigs_d[link[0][0]]
        for node in link[:-1]:
            down = contigs_d[node[1]]
            seq += down[node[11]:]
        # tail do not have link after itself
        if not is_circle:
            tail_seq = contigs_d[link[-1][1]]
            try:
                seq += tail_seq[link[-1][11]:]
            # rare
            except IndexError:
                pass
        else:
            seq = seq[:-link[-1][11]]
        seq.id = f'Merged_sequence {len(seq)}bp is_circle={is_circle}'
        seq.description = ''
        # seems circle is ok, non-circle is always bad result
        if is_circle:
            circular.append(seq)
        else:
            linear_d[len(seq)] = seq
    # keep longest
    linear_long = sorted(list(linear_d.values()), key=len,
                         reverse=True)[:max_linear]
    if len(circular) != 0:
        return circular
    else:
        log.warning('No circular assembly found.')
        return linear_long


def merge_main(arg_str=None):
    if arg_str is None:
        arg = parse_args()
    else:
        arg = parse_args(arg_str.split(' '))
    arg.input = [Path(i).absolute() for i in arg.input]
    if arg.out is None:
        # give another name?
        arg.out = arg.input[0].with_suffix('.merge')
    else:
        arg.out = Path(arg.out).absolute()
    contigs_and_rc, contigs_and_rc_fasta = get_contig(arg.input, arg.out)
    links = get_link(contigs_and_rc, contigs_and_rc_fasta)
    if len(links) == 0:
        merged = []
    else:
        merged = merge_seq(contigs_and_rc, links)
    if len(merged) == 0:
        log.critical('Failed to assembly contigs.')
    else:
        log.info(f'Got {len(merged)} assemblies.')
        SeqIO.write(merged, arg.out, 'fasta')
    contigs_and_rc_fasta.unlink()
    return len(merged), arg.out


if __name__ == '__main__':
    merge_main()



#!/usr/bin/python3

from logging import handlers
from pathlib import Path
import logging
import queue
import threading
try:
    from tkinter import messagebox, filedialog, scrolledtext
    import tkinter as tk
    from tkinter import ttk
except ImportError:
    print('Cannot find the module tkinter, please follow the README to install it.')
    print('Or try to use portable version https://github.com/wpwupingwp/novowrap/releases')
    print('Or, use conda instead (see README for details)')
    raise

from novowrap.utils import accessible
from novowrap.merge import merge_main
from novowrap.assembly import assembly_main
from novowrap.validate import validate_main


# define logger
# ui needn't coloredlogs
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
log = logging.getLogger('novowrap')


def scroll_text(window):
    """
    ScrolledText that shows logs.
    """
    def poll():
        while True:
            try:
                msg = log_queue.get(block=False)
                level = msg.levelname
                msg = formatter.format(msg) + '\n'
                scroll.insert('end', msg, level)
                scroll.yview('end')
            except queue.Empty:
                break
        # to avoid orphan poll()
        if log.hasHandlers():
            scroll.after(10, poll)
        else:
            return

    # clean old handlers
    for i in log.handlers:
        log.removeHandler(i)
    log_queue = queue.Queue()
    formatter = logging.Formatter(fmt=FMT, datefmt=DATEFMT)
    # do not add formatter to queuehandler, or msg will be formatted twice
    queue_handler = handlers.QueueHandler(log_queue)
    # give poll() time to quit
    root.after(100, log.addHandler(queue_handler))
    scroll = scrolledtext.ScrolledText(window)
    scroll.tag_config('INFO', foreground='black')
    scroll.tag_config('WARNING', foreground='orange')
    scroll.tag_config('ERROR', foreground='red')
    scroll.tag_config('CRITICAL', foreground='red')
    scroll.tag_config('EXCEPTION', foreground='red')
    scroll.pack(fill='both')
    scroll.after(0, poll)


def wlabel(window, text, row, column=0, width=25, padx=0, pady=0, sticky='ew',
           **kargs):
    """
    Generate and pack labels.
    """
    label = ttk.Label(window, text=text, width=width, anchor=tk.CENTER,
                      **kargs)
    label.grid(row=row, column=column, padx=padx, pady=pady, sticky=sticky)
    return label


def fentry(window, row, column, default='', padx=0, pady=0):
    """
    Generate and pack entrys.
    Fill with default string.
    """
    entry = ttk.Entry(window)
    entry.insert(0, default)
    entry.grid(row=row, column=column, padx=padx, pady=pady)
    return entry


def info(message):
    """
    For shorter.
    """
    messagebox.showinfo(message=message)


def after_close(frame):
    """
    Deiconify root before destroy.
    """
    def func():
        root.deiconify()
        frame.destroy()
    return func


def open_file(title, entry, single=True, entry2=None):
    """
    Set title, fill entry 1, empty entry 2.
    """
    def func():
        if single:
            a = filedialog.askopenfilename(title=title)
        else:
            a = filedialog.askopenfilenames(title=title)
        entry.delete(0, 'end')
        entry.insert(0, a)
        if entry2 is not None:
            entry2.delete(0, 'end')
    return func


def open_folder(title, entry):
    def func():
        a = filedialog.askdirectory(title=title)
        entry.delete(0, 'end')
        entry.insert(0, a)
    return func


def check_output(output, out_entry):
    """
    Make sure given path does not exist and writable.
    """
    output = Path(output).absolute()
    if output.exists():
        info('Output folder exists. Please use another folder.')
        out_entry.configure(style='hi.TButton')
        return False
    else:
        ok = accessible(output, 'folder')
        if not ok:
            info(f'You do not have permission to write in given output '
                 'folder. Please try another path.')
            out_entry.configure(style='hi.TButton')
            return False
    return True


def thread_wrap(function, arg_str, window):
    """
    Wrap for callback.
    The validate and merge share same return structure.
    Args:
        function(callable): function to call
        arg_str(str): string for fuction's argparse
        window(Toplevel): window to hide
    """
    try:
        result = function(arg_str)
    except Exception:
        log.exception('Abort.')
        info('Abort.')
        root.deiconify()
        return
    if result[0]:
        info(f'Done. See {result[1]} for details.')
    else:
        info(f'Fail. See {result[1]} for details.')
    window.withdraw()
    root.deiconify()
    return


def assembly_ui():
    """
    UI of assembly.
    """
    def show_adv():
        wroot.geometry(big_size)
        advance.grid(advance_info)

    def submit_assembly():
        # prepare arg_str
        arg_str = ''
        arg_input = input_entry.get()
        arg_input_list = list_entry.get()
        if arg_input and arg_input_list:
            info('Please only use one of "Input file" and "Input list"!')
            return
        elif arg_input == '' and arg_input_list == '':
            info('Input is required!')
            return
        # use dialog window to choose input, usually needn't check whether
        # exists
        elif arg_input:
            arg_str += f'-input {arg_input}'
        else:
            arg_str += f'-list {arg_input_list}'
        arg_ref = ref_entry.get()
        # use underscore in taxon name, which NCBI could handle
        arg_taxon = taxon_entry.get()
        if arg_ref and arg_taxon:
            info('Please only use one of "Genbank file" and "Taxonomy"!')
            return
        elif arg_ref == '' and len(arg_taxon) == 0:
            info('Reference is required!')
            return
        elif arg_ref != '' and len(arg_taxon) == 0:
            arg_str += f' -ref {arg_ref}'
        elif arg_ref == '' and len(arg_taxon) != 0:
            arg_str += f' -taxon {arg_taxon}'
        arg_out = out_entry.get()
        if arg_out == '"Current folder"':
            out_path = Path('.').absolute()
        else:
            out_path = Path(arg_out).absolute()
        if out_path.exists():
            if out_path.is_dir():
                out_path = out_path / 'Output'
            else:
                info('Invalid output folder name.')
                return
        if not check_output(out_path, out_entry):
            return
        arg_str += f' -out {out_path}'
        # advanced options
        arg_split = split_entry.get()
        if arg_split:
            arg_str += f' -split {arg_split}'
        arg_insert = insert_entry.get()
        if arg_insert:
            arg_insert += f' -insert_size {arg_insert}'
        arg_str += f' -platform {platform.get()}'
        arg_kmer = int(kmer_entry.get())
        if arg_kmer > 39 or arg_kmer < 23 or (arg_kmer % 2 != 1):
            info('K-mer should be an odd number in (23, 39)!')
            kmer_entry.configure(style='hi.TButton')
            return
        arg_str += f' -kmer {arg_kmer}'
        arg_size = size_entry.get()
        if '-' not in arg_size:
            info('Genome size should be "min-max" format!')
            size_entry.configure(style='hi.TButton')
            return
        min_size, max_size = arg_size.split('-')
        arg_str += f' -min {min_size} -max {max_size}'
        arg_seed = seed_entry.get().strip(',')
        if arg_seed:
            arg_str += f' -seed {arg_seed}'
        arg_seed_file = seed_file_entry.get()
        if arg_seed_file:
            arg_str += f' -seed_file {arg_seed_file}'
        # handle checkbutton
        if mt_mode.get():
            simple_validation.set(True)
            size_ok = messagebox.askyesno(
                title='Warning',
                message=(f'Given genome length range is {min_size}-{max_size}, '
                        f'ok for mitochondria?'))
            if not size_ok:
                return
            arg_str += ' -mt'
        else:
            if simple_validation.get():
                arg_str += ' -simple_validate'
        if simple_validation.get():
            info('The program will skip detecting quadripartite structure and '
                 'adjustment.')
        arg_s = float(s_entry.get())
        arg_l = float(l_entry.get())
        if max(arg_s, arg_l) > 1 or min(arg_s, arg_l) <= 0:
            info('Bad value!')
            s_entry.configure(style='hi.TButton')
            l_entry.configure(style='hi.TButton')
            return
        else:
            arg_str += f' -len_diff {arg_l} -perc_identity {arg_s}'
        # info(arg_str)
        # call validate
        wroot.withdraw()
        run = tk.Toplevel(root)
        run.geometry(size)
        run.title('Running...')
        run.wm_transient()
        frame = ttk.Frame(run)
        frame.pack(fill='both')
        scroll_text(frame)
        r = threading.Thread(target=thread_wrap,
                             args=(assembly_main, arg_str, run),
                             daemon=True)
        r.start()

    root.iconify()
    wroot = tk.Toplevel(root)
    wroot.geometry(size)
    wroot.title('Assembly')
    w = ttk.Frame(wroot)
    w.place(relx=0.5, rely=0.5, anchor='center')
    # use variable for easily edit
    row = 0
    inputs = ttk.LabelFrame(w, text='Input')
    inputs.grid(row=row, columnspan=4)
    row += 1
    wlabel(inputs, 'Input', row=row, column=1)
    input_entry = fentry(inputs, row=row, column=2)
    input_button = ttk.Button(inputs, text='Open', command=open_file(
        'Input file', input_entry, single=False))
    input_button.grid(row=row, column=3)
    row += 1
    # ttk label do not support 'fg'
    label1 = tk.Label(inputs, text='OR', fg='red')
    label1.grid(row=row, column=0, columnspan=2)
    row += 1
    wlabel(inputs, 'Input list', row=row, column=1)
    list_entry = fentry(inputs, row=row, column=2)
    list_button = ttk.Button(inputs, text='Open',
                             command=open_file('List file', list_entry,
                                               entry2=input_entry))
    list_button.grid(row=row, column=3)
    row += 1
    ref = ttk.LabelFrame(w, text='Reference')
    ref.grid(row=row, columnspan=4)
    row += 1
    wlabel(ref, 'Taxonomy', row=row, column=1)
    taxon_entry = fentry(ref, row=row, column=2, default='Nicotiana tabacum')
    row += 1
    label2 = tk.Label(ref, text='OR', fg='red')
    label2.grid(row=row, column=0, columnspan=2)
    row += 1
    wlabel(ref, 'Genbank file', row=row, column=1)
    ref_entry = fentry(ref, row=row, column=2)
    r_button = ttk.Button(ref, text='Open',
                          command=open_file('Reference file', ref_entry,
                                            entry2=taxon_entry))
    r_button.grid(row=row, column=3)
    row += 1
    wlabel(w, 'Output', row=row)
    out_entry = fentry(w, row=row, column=1, default=str(Path('.').absolute()))
    o_button = ttk.Button(w, text='Open',
                          command=open_folder('Output folder', out_entry))
    o_button.grid(row=row, column=2)
    row += 1

    advance = ttk.LabelFrame(w, text='Advanced settings')
    advance.grid(row=row, columnspan=3)
    advance_info = advance.grid_info()
    row += 1
    adv_input = ttk.LabelFrame(advance, text='Input')
    adv_input.grid(row=row)
    row += 1
    simple_validation = tk.BooleanVar()
    simple_validation.set(False)
    check1 = ttk.Checkbutton(adv_input, text='Simple validation',
                             variable=simple_validation, onvalue=True,
                             offvalue=False)
    check1.grid(row=row, column=1, sticky='w')
    row += 1
    mt_mode = tk.BooleanVar()
    mt_mode.set(False)
    check2 = ttk.Checkbutton(adv_input, text='Mitochondria genome',
                             variable=mt_mode, onvalue=True, offvalue=False)
    check2.grid(row=row, column=1, sticky='w')
    row += 1
    wlabel(adv_input, 'Split reads', row=row, column=0)
    split_entry = fentry(adv_input, row=row, column=1)
    row += 1
    wlabel(adv_input, 'Insert size', row=row, column=0)
    insert_entry = fentry(adv_input, row=row, column=1)
    row += 1
    platform = tk.StringVar()
    platform.set('illumina')
    wlabel(adv_input, 'Sequence platform', row=row, column=0)
    radio1 = ttk.Radiobutton(adv_input, text='illumina', variable=platform,
                             value='illumina')
    radio1.grid(row=row, column=1, sticky='w')
    radio2 = ttk.Radiobutton(adv_input, text='ion torrent', variable=platform,
                             value='ion')
    radio2.grid(row=row, column=2, sticky='w')
    row += 1
    adv_assembly = ttk.LabelFrame(advance, text='Assembly')
    adv_assembly.grid(row=row)
    row += 1
    wlabel(adv_assembly, 'K-mer (23-39, odd)', row=row, column=0)
    kmer_entry = fentry(adv_assembly, row=row, column=1, default='39')
    row += 1
    wlabel(adv_assembly, 'Genome Size range (bp)', row=row, column=0)
    size_entry = fentry(adv_assembly, row=row, column=1,
                        default='100000-200000')
    row += 1
    wlabel(adv_assembly, 'Seed genes', row=row, column=0)
    seed_entry = fentry(adv_assembly, row=row, column=1,
                        default='rbcL,psaB,psaC,rrn23')
    row += 1
    wlabel(adv_assembly, 'Seed file', row=row, column=0)
    seed_file_entry = fentry(adv_assembly, row=row, column=1)
    seed_button = ttk.Button(adv_assembly, text='Open',
                             command=open_file('Seed file', seed_file_entry,
                                               single=True))
    seed_button.grid(row=row, column=2)

    adv_validate = ttk.LabelFrame(advance, text='Validate')
    adv_validate.grid(row=row)
    row += 1
    wlabel(adv_validate, 'Sequence similarity (0-1)', row=row, column=0)
    s_entry = fentry(adv_validate, row=row, column=1, default='0.7')
    row += 1
    wlabel(adv_validate, 'Length difference (0-1)', row=row, column=0)
    l_entry = fentry(adv_validate, row=row, column=1, default='0.2')

    row += 1
    show_more = ttk.Button(w, text='More options', command=show_adv)
    show_more.grid(row=row, column=0, pady=10)
    ok = ttk.Button(w, text='Enter', command=submit_assembly)
    ok.grid(row=row, column=1, columnspan=2, sticky='EW', padx=40)
    advance.grid_forget()
    wroot.protocol('WM_DELETE_WINDOW', after_close(wroot))
    return


def merge_ui():
    """
    UI of merge.
    """
    def submit_merge():
        arg_str = ''
        inputs = input_entry.get()
        if inputs == '':
            info('Input is required!')
            return
        arg_str += f'-input {inputs}'
        arg_out = out_entry.get()
        if arg_out == '"Current folder"':
            out_path = Path('.').absolute()
        else:
            out_path = Path(arg_out).absolute()
        if out_path.exists():
            if out_path.is_dir():
                out_path = out_path / 'Output'
            else:
                info('Invalid output folder name.')
                return
        if not check_output(out_path, out_entry):
            return
        arg_str += f' -out {out_path}'
        print(arg_str)
        wroot.withdraw()
        run = tk.Toplevel(root)
        run.geometry(size)
        run.title('Running...')
        run.wm_transient()
        frame = ttk.Frame(run)
        frame.pack(fill='both')
        scroll_text(frame)
        r = threading.Thread(target=thread_wrap,
                             args=(merge_main, arg_str, wroot),
                             daemon=True)
        r.start()

    root.iconify()
    wroot = tk.Toplevel(root)
    wroot.geometry(small_size)
    wroot.title('Merge')
    frame = ttk.Frame(wroot)
    frame.place(relx=0.5, rely=0.5, anchor='center')
    row = 0
    wlabel(frame, 'Input', row=row, column=1)
    input_entry = fentry(frame, row=row, column=2)
    input_button = ttk.Button(frame, text='Open',
                              command=open_file('Input file', input_entry,
                                                single=False))
    input_button.grid(row=row, column=3)
    row += 1
    wlabel(frame, 'Output', row=row, column=1, pady=10)
    out_entry = fentry(frame, row=row, column=2, default='"Current folder"')
    o_button = ttk.Button(frame, text='Open',
                          command=open_folder('Output folder', out_entry))
    o_button.grid(row=row, column=3)
    row += 1
    ok = ttk.Button(frame, text='Enter', command=submit_merge)
    ok.grid(row=row, column=1, columnspan=3, sticky='EW', pady=10)
    wroot.protocol('WM_DELETE_WINDOW', after_close(wroot))
    return


def validate_ui():
    """
    UI of validate.
    """
    def submit_validate():
        # prepare arg_str
        arg_str = ''
        inputs = i_entry.get()
        if inputs == '':
            info('Input is required!')
            return
        else:
            arg_str += f'-input {inputs}'
        arg_ref = r_entry.get()
        # use underscore in taxon name, which NCBI could handle
        arg_taxon = t_entry.get().replace(' ', '_')
        if arg_ref and arg_taxon:
            info('Please only use one of "Reference file" and "Taxonomy"!')
            return
        elif arg_ref != '' and arg_taxon == '':
            arg_str += f' -ref {arg_ref}'
        elif arg_ref == '' and arg_taxon != '':
            arg_str += f' -taxon {arg_taxon}'
        arg_out = o_entry.get()
        if arg_out == '"Current folder"':
            out_path = Path('.').absolute()
        else:
            out_path = Path(arg_out).absolute()
        if out_path.exists():
            if out_path.is_dir():
                out_path = out_path / 'Output'
            else:
                info('Invalid output folder name.')
                return
        if not check_output(out_path, o_entry):
            return
        if mt_mode.get():
            simple_validation.set(True)
            arg_str += ' -mt'
        else:
            if simple_validation.get():
                arg_str += ' -simple_validate'
        if simple_validation.get():
            info('The program will skip detecting quadripartite structure and '
                 'adjustment.')
        # validate need existing out if called by others
        out_path.mkdir()
        out_tmp = out_path / 'Temp'
        out_tmp.mkdir()
        arg_str += f' -out {out_path}'
        arg_s = float(s_entry.get())
        arg_l = float(l_entry.get())
        if max(arg_s, arg_l) > 1 or min(arg_s, arg_l) <= 0:
            info('Bad value!')
            s_entry.configure(style='hi.TButton')
            l_entry.configure(style='hi.TButton')
            return
        else:
            arg_str += f' -len_diff {arg_l} -perc_identity {arg_s}'
        # call validate
        wroot.withdraw()
        run = tk.Toplevel(root)
        run.geometry(size)
        run.title('Running...')
        run.wm_transient()
        frame = ttk.Frame(run)
        frame.pack(fill='both')
        scroll_text(frame)
        r = threading.Thread(target=thread_wrap,
                             args=(validate_main, arg_str, run),
                             daemon=True)
        r.start()

    root.iconify()
    wroot = tk.Toplevel(root)
    wroot.geometry(size)
    wroot.title('Validate')
    # on top
    # wroot.wm_transient(root)
    w = ttk.Frame(wroot)
    w.place(relx=0.5, rely=0.5, anchor='center')
    # use variable for easily edit
    row = 0
    wlabel(w, 'Input', row=row, padx=15, pady=10)
    i_entry = fentry(w, row=row, column=1)
    i_button = ttk.Button(w, text='Open',
                          command=open_file('Input file', i_entry))
    i_button.grid(row=row, column=2)
    ref = ttk.LabelFrame(w, text='Reference')
    row += 1
    ref.grid(row=row, column=0, columnspan=5, padx=5, pady=8)
    row += 1
    wlabel(ref, 'Taxonomy', row=row, column=1)
    t_entry = fentry(ref, row=row, column=2, default='Nicotiana tabacum')
    row += 1
    label1 = tk.Label(ref, text='OR', fg='red')
    label1.grid(row=row, column=0, columnspan=2)
    row += 1
    wlabel(ref, 'Genbank/FASTA file', row=row, column=1, padx=13)
    r_entry = fentry(ref, row=row, column=2)
    r_button = ttk.Button(ref, text='Open',
                          command=open_file('Reference file (gb/fasta format)',
                                            r_entry, entry2=t_entry))
    r_button.grid(row=row, column=3)
    row += 1
    wlabel(w, 'Output', row=row, padx=10, pady=8)
    o_entry = fentry(w, row=row, column=1, default=str(Path('.').absolute()))
    o_button = ttk.Button(w, text='Open', command=open_folder('Output folder',
                                                              o_entry))
    o_button.grid(row=row, column=2)
    row += 1
    options = ttk.LabelFrame(w, text='Options')
    options.grid(row=row, padx=5, columnspan=5)
    row += 1
    simple_validation = tk.BooleanVar()
    simple_validation.set(False)
    check1 = ttk.Checkbutton(options, text='Simple validation',
                             variable=simple_validation, onvalue=True,
                             offvalue=False)
    check1.grid(row=row, column=1, sticky='w')
    row += 1
    mt_mode = tk.BooleanVar()
    mt_mode.set(False)
    check2 = ttk.Checkbutton(options, text='Mitochondria genome',
                             variable=mt_mode, onvalue=True, offvalue=False)
    check2.grid(row=row, column=1, sticky='w')
    row += 1
    wlabel(options, 'Sequence similarity (0-1)', row=row, padx=10)
    s_entry = fentry(options, row=row, column=1, default='0.7')
    row += 1
    wlabel(options, 'Length difference (0-1)', row=row, padx=10)
    l_entry = fentry(options, row=row, column=1, default='0.2')
    row += 1
    ok = ttk.Button(w, text='Enter', command=submit_validate)
    ok.grid(row=row, column=0, columnspan=3, sticky='EW', padx=50, pady=10)
    wroot.protocol('WM_DELETE_WINDOW', after_close(wroot))
    return


def ui_main():
    # init window
    global root
    root = tk.Tk()
    root.attributes('-topmost', 'true')
    w, h = root.winfo_screenwidth(), root.winfo_screenheight()
    s = min(w, h) // 2
    global size
    size = f'{s}x{int(s*0.618)}+{w//3}+{h//3}'
    global small_size
    small_size = f'{s}x{int(s*0.618/2)}+{w//3}+{h//3}'
    global big_size
    big_size = f'{s}x{int(s*0.618*2)}'
    style = ttk.Style()
    # don't know why 'hi.TEntry' can't work
    style.configure('hi.TButton', background='red')
    root.geometry(small_size)
    root.title('novowrap')
    # 1366x768
    if h < 800:
        root.tk.call('tk', 'scaling', 0.9)
    # 1440x900
    elif h < 1000:
        root.tk.call('tk', 'scaling', 0.95)
    # 2k
    elif h > 1400:
        root.tk.call('tk', 'scaling', 2.0)
    # cross-platform font?
    # default_font = font.nametofont('TkDefaultFont')
    # default_font_cfg = default_font.configure()
    # default_font.config(size=12)
    root_frame = ttk.Frame(root)
    # in center
    root_frame.place(relx=0.5, rely=0.5, anchor='center')
    # high dpi
    # root.tk.call('tk', 'scaling', 2.0)
    # btn->button
    assembly_btn = ttk.Button(root_frame, text='Assembly sequences',
                              command=assembly_ui)
    assembly_btn.grid(row=0, pady=10)
    validate_btn = ttk.Button(root_frame, text='Validate assemblies',
                              command=validate_ui)
    validate_btn.grid(row=1, pady=10)
    merge_button = ttk.Button(root_frame, text='Merge contigs',
                              command=merge_ui)
    merge_button.grid(row=2, pady=10)
    root.mainloop()


if __name__ == '__main__':
    ui_main()



#!/usr/bin/python3

from pathlib import Path
from shutil import unpack_archive
from subprocess import DEVNULL, run
from threading import Thread
from time import sleep
from urllib.request import urlopen
from zipfile import ZipFile
import gzip
import logging
import platform

from Bio import Entrez, SeqIO
from Bio.Alphabet import IUPAC
from Bio.Seq import reverse_complement as rc
from Bio.SeqFeature import SeqFeature, FeatureLocation


# define logger
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
log = logging.getLogger('novowrap')
try:
    import coloredlogs
    coloredlogs.install(level=logging.INFO, fmt=FMT, datefmt=DATEFMT)
except ImportError:
    pass
log = logging.getLogger('novowrap')


def accessible(name: Path, type_: str):
    """
    Check given path is accessible or not.
    Given path does not exist.
    Args:
        name(Path): folder or file, absolute path
        type_(str): 'folder' or 'file'
    Return:
        ok(bool): accessible or not
    """
    p = Path(name)
    if type_ == 'folder':
        try:
            p.mkdir()
            p.rmdir()
            ok = True
        except PermissionError:
            ok = False
    elif type_ == 'file':
        try:
            p.touch()
            p.unlink()
            ok = True
        except PermissionError:
            ok = False
    else:
        log.critical(f'Illegal type: {type_}')
        ok = False
    return ok


def move(source: Path, dest, copy=False):
    """
    Move source to dest and return dest.
    If set "copy", copy source to dest instead of move.
    Because Path.rename could not move file across different filesystem or
    drive, have to use copy and delete to implement "move".
    Warning:
        This function does not check whether dest exists or not.
    Args:
        source(Path): old path
        dest(Path or str): new path
        copy(bool): copy or move
    Return:
        dest(Path): new path
    """
    source = Path(source).absolute()
    dest = Path(dest).absolute()
    # avoid useless copy
    # Path.samefile may raise FileNotFoundError
    if source == dest:
        log.debug(f'{source} and {dest} are same.')
    else:
        # read_bytes/write_bytes includes open, read/write and close steps
        dest.write_bytes(source.read_bytes())
        if not copy:
            source.unlink()
    return dest


def get_full_taxon(taxon: str):
    """
    Get full lineage of given taxon, return lineage list.
    Not only contains Kingdom, Phylum, Class, Order, Family, Genus, Species.
    Arg:
        taxon(str): given taxon name, could be common name
    Return:
        ok(bool): query ok or not
        lineage(list): lineage list, from lower rank to higher, empty for fail
    """
    # make copy
    name = taxon.capitalize()
    try:
        search = Entrez.read(Entrez.esearch(db='taxonomy', term=f'"{name}"'))
    except Exception:
        return False, []
    if search['Count'] == '0':
        if ' ' not in name:
            return False, []
        if ' ' in name:
            name = name.split(' ')[0]
            sleep(0.5)
            try:
                search = Entrez.read(Entrez.esearch(db='taxonomy',
                                                    term=f'"{name}"'))
            except Exception:
                return [], False
            if search['Count'] == '0':
                return False, []
    taxon_id = search['IdList'][0]
    try:
        record = Entrez.read(Entrez.efetch(db='taxonomy', id=taxon_id))[0]
    except Exception:
        return False, []
    # names = [i['ScientificName'] for i in record['LineageEx']]
    full_lineage = [(i['Rank'], i['ScientificName']) for i in
                    record['LineageEx']]
    full_lineage.append((record['Rank'], record['ScientificName']))
    return True, reversed(full_lineage)


def get_ref(taxon: str, out: Path, tmp=None, mt_mode=False,
            simple_validate=False):
    """
    Get reference gb file.
    Only one record will be retrieved.
    Arg:
        taxon(str): given taxon name
        out(Path): output folder
        tmp(None or Path): temp folder
    Return:
        ref(Path or None): gb file, None for fail
        ref_taxon(str or None): taxon of reference's, may not be same with
        given taxon
    """
    log.info(f'Try to get reference of {taxon} from NCBI Genbank.')
    Entrez.email = 'guest@example.org'
    # handle quotation mark
    taxon = taxon.strip('"')
    taxon = taxon.strip("'")
    try:
        Entrez.read(Entrez.esummary(db='taxonomy', id='9606'))
    except Exception:
        log.critical('Failed to query on NCBI Entrez server, please check '
                     'your Internet connection.')
        return None, None
    if tmp is None:
        tmp = out
    get_taxon_ok = False
    lineage = []
    for try_ in range(3):
        get_taxon_ok, lineage = get_full_taxon(taxon)
        if get_taxon_ok:
            break
    if not get_taxon_ok:
        log.critical(f'Cannot find {taxon} in NCBI Taxonomy database.')
        log.warning("Please check the taxon name's spell.")
        return None, None
    for taxon in lineage:
        rank, taxon_name = taxon
        if taxon_name == '':
            continue
        if rank == 'order':
            log.critical('The taxonomy of the reference is not close-related.')
            log.critical('The result may be incorrect.')
        taxon_name = taxon_name.replace(' ', '_')
        # Entrez has limitation on query frequency (3 times per second)
        # https://www.ncbi.nlm.nih.gov/books/NBK25497/#chapter2.Usage_Guidelines_and_Requiremen
        sleep(0.5)
        if mt_mode:
            query = (f'''{taxon_name}[Organism] AND refseq[filter] '''
                     f'''AND mitochondrion[filter]''')
        else:
            query = (f'''{taxon_name}[Organism] AND refseq[filter] '''
                     f'''AND (chloroplast[filter] OR plastid[filter])''')
        log.debug(f'Query from NCBI Genbank:\t{query}')
        # seems nuccore is more stable than taxonomy database
        handle = Entrez.read(Entrez.esearch(db='nuccore', term=query,
                                            usehistory='y'))
        count = int(handle['Count'])
        if count == 0:
            continue
        info = Entrez.read(Entrez.esummary(db='nuccore',
                                           webenv=handle['WebEnv'],
                                           query_key=handle['QueryKey']))
        accession = info[0]['Caption']
        ref = out / f'{taxon_name}_{accession}.gb'
        content = Entrez.efetch(db='nuccore', webenv=handle['WebEnv'],
                                query_key=handle['QueryKey'], rettype='gb',
                                retmode='text', retmax=1)
        with open(ref, 'w') as _:
            _.write(content.read())
        r_gb, r_fasta = rotate_seq(ref, tmp=tmp,
                                   simple_validate=simple_validate)
        if r_gb is not None:
            r_gb.unlink()
            r_fasta.unlink()
            log.info(f'Got {ref.name} as reference.')
            return ref, taxon_name
        else:
            continue
    return None, None


def blast(query: Path, target: Path, perc_identity=70):
    """
    Use simple BLAST with special output format.
    Args:
        query(Path): query filename
        target(Path): target filename
        perc_identity(float): perc_identity for BLAST
    Return:
        blast_out(Path): blast result filename
        blast_log(Path): blast log

    """
    fmt = ('qseqid sseqid sstrand qlen slen length pident gapopen qstart qend '
           'sstart send')
    query = query.absolute()
    target = target.absolute()
    blast_out = query.with_suffix('.blast')
    blast_log = query.with_suffix('.blast_log')
    # use blastn -subject instead of makeblastdb
    ok, blastn = get_blast()
    if not ok:
        log.critical('Failed to run BLAST.')
        return None, None
    b_log = open(blast_log, 'w')
    b_run = run(f'{blastn} -query {query} -subject {target} -outfmt "7 {fmt}" '
                f'-out {blast_out} -strand both -perc_identity '
                f'{perc_identity}',
                shell=True, stdout=b_log, stderr=b_log)
    b_log.close()
    # remove makeblastdb result
    if b_run.returncode != 0:
        log.critical('Cannot run BLAST.')
        return None, blast_log
    return blast_out, blast_log


def parse_blast_tab(filename: Path):
    """
    Parse BLAST result (tab format).
    Return [qseqid, sseqid, sstrand, qlen, slen, length, pident, gapopen,
    qstart, qend, sstart, send]
    Arg:
        filename(Path): blast result file
    Return:
        line(list): parsed result
    """
    query = []
    # empty for empty
    if filename is None:
        return []
    # blastn use system's default encoding
    with open(filename, 'r') as raw:
        for line in raw:
            if line.startswith('# BLAST'):
                if query:
                    yield query
                query.clear()
            elif line.startswith('#'):
                pass
            else:
                line = line.strip().split('\t')
                # last is str
                line[3:] = [int(float(i)) for i in line[3:]]
                query.append(line)


def repeat(filename, fmt):
    """
    Duplicate sequence to get full length of IR which may locate in start or
    end of sequences that BLAST cannot get whole length.
    Args:
        filename(Path): file to be repeat
        fmt(str): 'gb' or 'fasta'
    Return:
        new_file(Path): repeated fasta
    """
    new_file = filename.with_suffix('.rpt')
    raw = SeqIO.read(filename, fmt)
    # assume given sequence is a whole chloroplast genome, no more or less
    new = raw + raw
    log.debug(f'Before {len(raw)}; Repeat {len(new)}')
    SeqIO.write(new, new_file, 'fasta')
    return new_file


def slice_gb(seq, location):
    """
    Biopython will skip incomplete annotation.
    Use this function to keep those fragments.
    Arg:
        seq(SeqRecord): sequence
        location(FeatureLocation): location
    Return:
        new_seq(SeqRecord): new SeqRecord
    """
    start = location.start
    end = location.stop
    f_before = None
    f_after = None
    for f in seq.features:
        if start in f:
            # slice may be inner of a feature
            f_before = SeqFeature(id=f.id, type='part',
                                  location=FeatureLocation(
                                      start-start,
                                      min(f.location.end, end)-start,
                                      f.location.strand),
                                  qualifiers=f.qualifiers)
            f_before.qualifiers['raw_location'] = str(f.location)
            break
    for f in seq.features:
        if end in f:
            f_after = SeqFeature(id=f.id, type='part',
                                 location=FeatureLocation(
                                     max(f.location.start, start)-start,
                                     end-start,
                                     f.location.strand),
                                 qualifiers=f.qualifiers)
            f_after.qualifiers['raw_location'] = str(f.location)
            break
    # both are not None
    if all([f_before, f_after]):
        if (f_before.location == f_after.location and
                f_before.qualifiers == f_after.qualifiers):
            # remove duplicated feature
            f_after = None
    new_seq = seq[location]
    for i in (f_before, f_after):
        if i is not None:
            i.qualifiers['sliced'] = str(location)
            i.qualifiers['reference'] = seq.id
            new_seq.features.append(i)
    return new_seq


def get_fmt(filename):
    """
    Detect file format.
    Support gz, fasta and gb.
    Args:
        filename(Path or str): filename
    Return:
        fmt(str): 'fasta', 'gz', 'gb' or 'txt' (others)
    """
    with gzip.open(filename) as _:
        try:
            _.readline()
            return 'gz'
        except OSError:
            pass
    with open(filename, 'r') as _:
        peek = _.readline()
        if peek.startswith('>'):
            fmt = 'fasta'
        elif peek.upper().startswith('LOCUS'):
            fmt = 'gb'
        else:
            fmt = 'txt'
    return fmt


def rotate_seq(filename, min_ir=1000, tmp=None, silence=True,
               simple_validate=False, retry=False):
    """
    Rotate genbank or fasta record, from LSC (trnH-psbA) to IRa, SSC, IRb.
    Input file should only contains one record.
    Arg:
        filename(Path): genbank or filename
        min_IR: minimum IR length
        tmp(Path or None): temporary folder
        silence(bool): print debug info or not
        simple_validate(bool): use abnormal rotate or not
        retry(bool): retry with less restrictions
    Return:
        new_gb(Path): gb file
        new_fasta(Path): fasta file
    """
    # normal rotate
    # allow 50 bases difference
    _allowed_difference = 50
    filename = Path(filename).absolute()
    if tmp is None:
        tmp = filename.parent
    if silence:
        log.setLevel(logging.CRITICAL)
    log.info(f'Rotate {filename}...')
    fmt = get_fmt(filename)
    # get origin seq
    origin_seq = list(SeqIO.parse(filename, fmt))
    assert len(origin_seq) == 1
    origin_seq = origin_seq[0]
    origin_seq.seq.alphabet = IUPAC.ambiguous_dna
    origin_len = len(origin_seq)
    # get repeat seq
    repeat_fasta = repeat(filename, fmt)
    repeat_seq = SeqIO.read(repeat_fasta, 'fasta')
    log.setLevel(logging.INFO)
    blast_result, blast_log = blast(repeat_fasta, repeat_fasta)
    log.setLevel(logging.CRITICAL)
    if blast_result is None:
        log.setLevel(logging.INFO)
        return None, None
    # clean tmp files immediately, in case of exceptions that break clean step
    repeat_fasta.unlink()
    new_fasta = filename.with_suffix('.rotate')
    if filename.suffix == '.gb':
        new_gb = filename.with_suffix('.gb.gb')
    else:
        new_gb = filename.with_suffix('.gb')
    if simple_validate:
        # do not rotate
        SeqIO.write(origin_seq, new_fasta, 'fasta')
        SeqIO.write(origin_seq, new_gb, 'gb')
        log.debug('simple rotate completed.')
        log.setLevel(logging.INFO)
        return new_gb, new_fasta
    success = False
    # only one record, loop just for for unpack
    for query in parse_blast_tab(blast_result):
        locations = set()
        # use fasta's description
        # for the second round
        ambiguous_base_n = len(str(origin_seq.seq).strip('ATCGatcg'))
        if retry:
            ambiguous_base_n = max(ambiguous_base_n, _allowed_difference)
            log.info(f'Allowing {ambiguous_base_n} bases difference this time.')
        # only use hit of IR to IR
        max_aln_n = 0
        # because of ambiguous base, percent of identity bases may not be 100%
        if ambiguous_base_n != 0:
            log.warning(f'\tFound {ambiguous_base_n} ambiguous bases.')
            p_ident_min = int((1-(ambiguous_base_n/origin_len))*100)
        else:
            p_ident_min = 100
        for hit in query:
            (qseqid, sseqid, sstrand, qlen, slen, length, pident, gapopen,
             qstart, qend, sstart, send) = hit
            # name = seq.name
            # only self
            if qseqid != sseqid:
                continue
            # skip too short match
            if length < min_ir:
                continue
            # too long, origin to repeat
            if length >= origin_len:
                continue
            # allow few gaps
            if gapopen != 0:
                if gapopen > ambiguous_base_n:
                    continue
            if pident < p_ident_min:
                continue
            # mismatch
            location = tuple(sorted([qstart, qend, sstart, send]))
            # hit across origin and repeat
            if location[-1] - location[0] > origin_len:
                continue
            # self to self or self to repeat self
            if len(set(location)) != 4:
                continue
            # filter short hit
            # the longest is IR
            if length < max_aln_n:
                continue
            else:
                max_aln_n = length
            locations.add(location)
        if not locations:
            continue
        locations = list(locations)
        locations.sort(key=lambda x: x[1]-x[0], reverse=True)
        locations.sort(key=lambda x: x[0])
        if retry:
            _ = []
            last = -1
            for i in locations:
                if i[0] == last:
                    continue
                else:
                    _.append(i)
                    last = i[0]
            locations = list(_)
        if len(locations) < 2:
            continue
        if len(locations) < 2:
            continue
        # remove extra hit, first two if enough
        locations = locations[:2]
        # ira_start, ira_end, irb_start, irb_end
        a = slice(locations[0][0]-1, locations[0][1])
        b = slice(locations[0][1], locations[0][2]-1)
        c = slice(locations[0][2]-1, locations[0][3])
        d = slice(locations[1][1], locations[1][2]-1)
        if (locations[0][2]-locations[0][1]) > (
                locations[1][2]-locations[1][1]):
            region_LSC = b
            region_IRa = c
            region_SSC = d
            region_IRb = a
        else:
            region_LSC = d
            region_IRa = a
            region_SSC = b
            region_IRb = c
        # self to self is the longest match
        seq_LSC = slice_gb(repeat_seq, region_LSC)
        seq_IRa = slice_gb(repeat_seq, region_IRa)
        seq_SSC = slice_gb(repeat_seq, region_SSC)
        seq_IRb = slice_gb(repeat_seq, region_IRb)
        new_seq = seq_LSC + seq_IRa + seq_SSC + seq_IRb
        # to be continue
        new_seq.seq.alphabet = IUPAC.ambiguous_dna
        if origin_len != len(new_seq):
            log.warning(f'\tOld and new sequences do not have save length.')
            log.info(f'Old: {origin_len}\tNew: {len(new_seq)}')
            if abs(origin_len - len(new_seq)) > ambiguous_base_n:
                log.info(f'\tToo much difference. Reject.')
                continue
        if len(seq_IRa) != len(seq_IRb):
            log.warning(f'\tIRa ({len(seq_IRa)}) and IRb ({len(seq_IRb)}) do '
                        'not have same length.')
            if abs(len(seq_IRa) - len(seq_IRb)) > ambiguous_base_n:
                log.warning(f'\tToo much difference. Reject.')
                continue
        # output
        offset = -1
        features = []
        for f_name, f in zip(('large single copy (LSC)',
                              'inverted repeat A (IRa)',
                              'small single copy (SSC)',
                              'inverted repeat B (IRb)'),
                             (region_LSC, region_IRa, region_SSC, region_IRb)):
            length = f.stop - f.start
            features.append(SeqFeature(
                FeatureLocation(offset+1, length+offset+1),
                type='misc_feature',
                qualifiers={'note': f_name, 'software': 'rotate_seq'},
                strand=1))
            offset += length
        new_seq.features.extend(features)
        # print feature
        log.info(f'\tRegions of rotated record:')
        for f in features:
            log.info(f'\t{f.qualifiers["note"][-4:-1]}: {f.location.start} to '
                     f'{f.location.end} ({len(f)} bp)')
        SeqIO.write(new_seq, new_gb, 'gb')
        SeqIO.write(new_seq, new_fasta, 'fasta')
        success = True
    blast_result.unlink()
    blast_log.unlink()
    log.setLevel(logging.INFO)
    if not success:
        log.debug('Retry with less restrictions...')
        if retry:
            log.critical(f'Failed to rotate {filename}.')
            return None, None
        else:
            return rotate_seq(filename, min_ir, tmp, silence, simple_validate,
                              retry=True)
    return new_gb, new_fasta


def get_regions(gb):
    """
    Arg:
        gb(Path): rotate_seq generated gb file, only contains one record
    Return:
        region({name: SeqFeature}): region location dict, default is ''
    """
    ref_region = dict().fromkeys(['LSC', 'SSC', 'IRa', 'IRb'], '')
    for feature in SeqIO.read(gb, 'gb').features:
        if (feature.type == 'misc_feature' and
                feature.qualifiers.get('software', ['', ])[0] == 'rotate_seq'):
            key = feature.qualifiers['note'][0][-4:-1]
            value = feature
            ref_region[key] = value
    return ref_region


def rc_regions(gb, choice='whole'):
    """
    Reverse and complement given region of sequence.
    Args:
        gb(Path or str): rotate_seq generated gb file
        choice(str): region to be processed, must be in 'LSC', 'IRa', 'SSC',
        'IRb', 'whole'.
    Return:
        new_file(Path): reverse-complemented fasta
    """
    # choices = ('LSC', 'IRa', 'SSC', 'IRb', 'whole')
    raw = SeqIO.read(gb, 'gb')
    data = {}
    new_seq = ''
    regions = get_regions(gb)
    for r in regions:
        data[r] = regions[r].extract(raw).seq
    if choice != 'whole':
        data[choice] = rc(regions[choice].extract(raw.seq))
        new_seq = data['LSC']
        for i in ['IRa', 'SSC', 'IRb']:
            new_seq += data[i]
    else:
        new_seq = rc(raw.seq)
    new_name = '_RC_' + raw.name
    new_file = gb.with_suffix('.rc.rc')
    with open(new_file, 'w') as _:
        _.write(f'>{new_name}\n')
        _.write(f'{new_seq}\n')
    return new_file


def test_cmd(program, option='-v'):
    """
    Test given program and option is ok to run or not.
    Args:
        program(Path or str): program path, could be relative path if it can
        be found in $PATH or %PATH%
        option(str): option for program, usually use "-v" to show version to
        test the program
    Return:
        success(bool): success or not
    """
    test = run(f'{program} {option}', shell=True, stdout=DEVNULL,
               stderr=DEVNULL)
    success = True if test.returncode == 0 else False
    return success


def get_third_party():
    """
    Get third_party folder.
    If do not exist, create it.
    If cannot access, report.
    Return:
        success(bool): ok or not
        third_party(Path): absolute path of third_party folder
    """
    third_party = Path().home().absolute() / '.novowrap'
    success = False
    if not third_party.exists():
        log.debug(f'Create folder {third_party}')
        try:
            third_party.mkdir()
        except Exception:
            log.critical(f'Failed to create {third_party}.'
                         'Please contact the administrator.')
            return success, third_party
    success = True
    return success, third_party


def get_perl(third_party=None):
    """
    Linux and Mac have perl already.
    For Windows user, this function help to get perl.exe .
    Only support x86_64 or amd64 machine.
    Args:
        third_party(Path or None): path for install
    Return:
        perl(str): perl location, empty for fail
    """
    url = ('http://strawberryperl.com/download/5.30.1.1/'
           'strawberry-perl-5.30.1.1-64bit-portable.zip')
    # have to test for threading
    if third_party is None:
        third_party_ok, third_party = get_third_party()
        if not third_party_ok:
            return ''
    # only Windows need it
    home_perl = third_party / 'strawberry_perl' / 'perl' / 'bin' / 'perl.exe'
    # use run instead of find_executable because the later only check if exist
    # and ignore if could run
    perl = 'perl'
    if test_cmd(perl):
        return perl
    if test_cmd(home_perl):
        return str(home_perl)
    if platform.system() != 'Windows':
        log.critical('Cannot find Perl. Please follow the link '
                     'https://www.perl.org/get.html to install.')
        return ''
    else:
        log.warning('Cannot find Perl. Try to install.')
        if '64' not in platform.machine():
            log.critical(f'Unsupport machine {platform.machine()}')
            return ''
        try:
            # file is 148mb, 148mb/3000s=~50kb/s, consider it's ok for
            # most of users
            log.info('Download may be slow. Please consider manualy install '
                     'Perl according to https://www.perl.org/get.html ')
            down = urlopen(url, timeout=3000)
        except Exception:
            log.critical('Cannot download Perl. Please try to manually '
                         ' install.')
            return ''
        zip_file = third_party / 'strawberry-perl.zip'
        with open(zip_file, 'wb') as out:
            out.write(down.read())
        folder = third_party / 'strawberry_perl'
        try:
            with ZipFile(zip_file) as z:
                z.extractall(folder)
        except Exception:
            log.critical('The file is damaged.')
            log.critical('Please check your Internet connection.')
            return ''
        # fixed path in zip file, should not be wrong
        assert test_cmd(home_perl)
        return str(home_perl)


def get_novoplasty(third_party=None):
    """
    Ensure perl and novoplasty is available.
    Return novoplasty's path or None.
    Args:
        third_party(Path or None): path for install
    Return:
        novoplasty(Path or None): path of perl file, None for fail
    """
    url = 'https://github.com/ndierckx/NOVOPlasty/archive/NOVOPlasty3.7.2.zip'
    if third_party is None:
        third_party_ok, third_party = get_third_party()
        if not third_party_ok:
            return None
    novoplasty = (third_party / 'NOVOPlasty-NOVOPlasty3.7.2' /
                  'NOVOPlasty3.7.2.pl')
    if novoplasty.exists():
        log.debug('Found NOVOPlasty in third_party folder.')
        return novoplasty
    log.critical('Cannot find NOVOPlasty, try to download.')
    # license warning
    log.info('\tThe program assumes that user ACCEPT the license of '
             'NOVOPlasty. ')
    log.info('\tIF NOT, please terminate the running.')
    log.info('\tSee https://raw.githubusercontent.com/'
             'ndierckx/NOVOPlasty/ for details.')
    try:
        # 126kb/10s=12.6kb/s, enough for test
        _ = urlopen('https://github.com', timeout=10)
    except Exception:
        log.critical('Cannot connect to github.com.')
        log.critical('Please check your Internet connection.')
        return None
    log.info('Due to connection speed, may need minutes. Please be patient.')
    try:
        # file is ~18mb, 18mb/360s=50kb/s, consider it's ok for most of users
        down = urlopen(url, timeout=360)
    except Exception:
        log.critical('Cannot download NOVOPlasty.')
        log.critical('Please manually download it from '
                     'https://github.com/ndierckx/NOVOPlasty')
        return None
    zip_file = third_party / 'NOVOPlasty3.7.2.zip'
    with open(zip_file, 'wb') as out:
        out.write(down.read())
    # novoplasty's files have illegal character ":" which cannot be extract by
    # shutil.unpack_archive in Windows, have to use zipfile
    # windows and linux both use "/"
    try:
        with ZipFile(zip_file, 'r') as z:
            z.extractall(third_party)
    except Exception:
        log.critical('The file is damaged.')
        log.critical('Please check your Internet connection.')
        return None
    log.info(f'Got {novoplasty.stem}.')
    return novoplasty


def get_blast(third_party=None):
    """
    Get BLAST location.
    If BLAST was found, assume makeblastdb is found, too.
    If not found, download it.
    Args:
        third_party(Path or None): path for install
    Return:
        ok(bool): success or not
        blast(str): blast path
    """
    if third_party is None:
        third_party_ok, third_party = get_third_party()
        if not third_party_ok:
            return third_party_ok, ''
    home_blast = third_party / 'ncbi-blast-2.10.0+' / 'bin' / 'blastn'
    # in Windows, ".exe" can be omitted
    # win_home_blast = home_blast.with_name('blastn.exe')
    ok = False
    # older than 2.8.1 is buggy
    url = ('ftp://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/2.10.0/'
           'ncbi-blast-2.10.0+')
    urls = {'Linux': url+'-x64-linux.tar.gz',
            'Darwin': url+'-x64-macosx.tar.gz',
            'Windows': url+'-x64-win64.tar.gz'}
    blast = 'blastn'
    if test_cmd(blast, '-version'):
        ok = True
        return ok, blast
    if test_cmd(home_blast, '-version'):
        ok = True
        return ok, str(home_blast)
    log.warning('Cannot find NCBI BLAST, try to install.')
    log.info('According to Internet speed, may be slow.')
    try:
        # 50kb/10s=5kb/s, enough for test
        _ = urlopen('https://www.ncbi.nlm.nih.gov', timeout=10)
    except Exception:
        log.critical('Cannot connect to NCBI.')
        log.critical('Please check your Internet connection.')
        return ok, ''
    try:
        # file is 86-222mb, 222mb/3600s=60kb/s, consider it's ok for users all
        # over the world
        down = urlopen(urls[platform.system()], timeout=3600)
    except Exception:
        log.critical('Cannot download BLAST.')
        log.critical('Please manually download it from'
                     'ftp://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/')
        return ok, ''
    down_file = third_party / 'BLAST_2.10.0.tar.gz'
    with open(down_file, 'wb') as out:
        out.write(down.read())
    try:
        unpack_archive(down_file, third_party)
    except Exception:
        log.critical('The file is damaged.')
        log.critical('Please check your connection.')
        return ok, ''
    assert test_cmd(home_blast, '-version')
    ok = True
    return ok, str(home_blast)


def get_all_third_party():
    """
    Use three threads to speed up.
    """
    log.info('Try to locate or install all third-party software.')
    third_party_ok, third_party = get_third_party()
    if not third_party_ok:
        return -1
    perl = Thread(target=get_perl, args=(third_party,), daemon=True)
    novoplasty = Thread(target=get_novoplasty, args=(third_party,),
                        daemon=True)
    blast = Thread(target=get_blast, args=(third_party,), daemon=True)
    perl.start()
    novoplasty.start()
    blast.start()
    perl.join()
    novoplasty.join()
    blast.join()
    log.info('Finished.')


if __name__ == '__main__':
    get_all_third_party()



#!/usr/bin/python3

import setuptools

with open('README.md', 'r') as _:
    long_description = _.read()

with open('requirements.txt', 'r') as _:
    requires = [i.strip() for i in _.readlines()]

setuptools.setup(
    author='Ping Wu',
    author_email='wpwupingwp@outlook.com',
    description='Assembly Plastid Genome',
    install_requires=requires,
    license='GNU AGPL v3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='novowrap',
    packages=setuptools.find_packages(),
    # f-string and Path-like
    python_requires='>=3.7',
    url='https://github.com/wpwupingwp/novowrap',
    version='1.1',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
)



#!/usr/bin/python3

from os import chdir
from pathlib import Path
from random import randint
from subprocess import DEVNULL, run
from threading import Thread
from time import sleep
import argparse
import gzip
import logging
import platform

from Bio import SeqIO

from novowrap import utils
from novowrap.merge import merge_main
from novowrap.validate import validate_main


# define logger
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
log = logging.getLogger('novowrap')
try:
    import coloredlogs
    coloredlogs.install(level=logging.INFO, fmt=FMT, datefmt=DATEFMT)
except ImportError:
    pass


def parse_args(arg_list=None):
    arg = argparse.ArgumentParser(
        prog='assembly.py',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    inputs = arg.add_argument_group('Input')
    inputs.add_argument('-input', nargs='*', required=True,
                        help='single or pair-end input, fastq or gz format')
    inputs.add_argument('-list', dest='list_file',
                        help='csv file for batch mode')
    inputs.add_argument('-platform', choices=['illumina', 'ion'],
                        default='illumina', help='sequencing platform')
    inputs.add_argument('-insert_size', type=int,
                        help='insert size of sequencing library')
    inputs.add_argument('-seed', default='rbcL,psaB,psaC,rrn23',
                        help='seed gene, separated by comma')
    inputs.add_argument('-seed_file',
                        help='seed file, will overwrite "-seed" option')
    inputs.add_argument('-split', default=0, type=int,
                        help='reads to use, set to 0 to skip split')
    options = arg.add_argument_group('Option')
    options.add_argument('-kmer', choices=range(23, 40, 2), default=39,
                         type=int, help='kmer size')
    options.add_argument('-min', default=100000, type=int,
                         help='minimum genome size (KB)')
    options.add_argument('-max', default=200000, type=int,
                         help='maximum genome size (KB)')
    options.add_argument('-mem', default=30, type=int,
                         help='maximum memory (GB)')
    options.add_argument('-out', help='output folder')
    options.add_argument('-debug', action='store_true', help='debug mode')
    options.add_argument('-simple_validate', action='store_true',
                         help='for plastids with abnormal structure')
    options.add_argument('-mt', dest='mt_mode', action='store_true',
                         help='for mitochondria (EXPERIMENTAL)')
    reference = arg.add_argument_group('Reference')
    reference.add_argument('-ref',
                           help='reference file, should be "gb" format with '
                           'only one record')
    reference.add_argument('-taxon', nargs='*',
                           default=['Nicotiana', 'tabacum'],
                           help='Taxonomy name')
    validate = arg.add_argument_group('Validate')
    validate.add_argument('-perc_identity', type=float, default=0.7,
                          help='minimum percentage of identity of BLAST, 0-1')
    validate.add_argument('-len_diff', type=float, default=0.2,
                          help='maximum percentage of length differnce of '
                          'query to reference, 0-1')
    if arg_list is None:
        return arg.parse_args()
    else:
        return arg.parse_args(arg_list)


def _get_name(inputs: list):
    """
    Given list of input, return output file's name.
    """
    out_name = Path('Output').absolute()
    if len(inputs) == 1:
        # for single file, directly remove all suffixes is dangerous
        out_name = Path(f'{Path(inputs[0]).stem}-out').absolute()
        return out_name
    f = Path(inputs[0])
    r = Path(inputs[1])
    while f.suffix == r.suffix and f.suffix != '':
        f = f.with_suffix('')
        r = r.with_suffix('')
    same = 0
    idx = 0
    for i, j in zip(str(f), str(r)):
        if i == j:
            same += 1
        else:
            break
        idx += 1
    if same != 0:
        _ = list(str(f))
        _.pop(idx)
        strip_ = ''.join(_).rstrip('-_')
        if len(strip_) != 0:
            out_name = Path(strip_).absolute()
    return out_name


def get_output(arg):
    """
    Get output folder.
    If exists, return None.
    Args:
        arg(NameSpace): arg generated by parse_args()
    Return:
        out(Path or None): output path
    """
    out = Path('.').absolute()
    if arg.out is None:
        if arg.input is not None:
            out = _get_name(arg.input)
        elif arg.list_file is not None:
            out = Path(Path(arg.list_file).stem).absolute()
        else:
            raise ValueError('This should not happen.')
    else:
        out = Path(arg.out).absolute()
    if out.exists():
        log.warning(f'Output folder {out.name} exists.')
        # give users one more chance
        new_name = out.name + f'-{randint(0, 2020):04d}'
        out = out.with_name(new_name)
        if out.exists():
            log.critical('Cannot create output folder.')
            return None
        else:
            log.info(f'Use {out.name} instead.')
    return out.absolute()


def init_arg(arg):
    """
    Initialize working folder with arg.
    Args:
        arg(NameSpace): arg generated from parse_args
    Return:
        success(bool): success or not
        arg(NameSpace): arg with output info
    """
    if arg.debug:
        logging.basicConfig(level=logging.DEBUG)
        try:
            import coloredlogs
            coloredlogs.install(level=logging.DEBUG, fmt=FMT, datefmt=DATEFMT)
        except ImportError:
            pass
    success = False
    if arg.list_file is None and arg.input is None:
        log.critical('Input is empty.')
        return success, arg
    elif arg.list_file is not None and arg.input is not None:
        log.critical('Cannot use both "-input" and "-list".')
        return success, arg
    elif arg.list_file is not None:
        arg.list_file = Path(arg.list_file).absolute()
        if not arg.list_file.exists():
            log.critical(f'Input file {arg.list_file} does not exists.')
            return success, arg
    else:
        if len(arg.input) > 2:
            log.critical('Only accept one or two input file(s).')
            return success, arg
        arg.input = [Path(i).absolute() for i in arg.input]
        for i in arg.input:
            if not i.exists():
                log.critical(f'Input file {i} does not exists.')
                return success, arg
    if arg.ref is not None:
        arg.ref = Path(arg.ref).absolute()
        if not arg.ref.exists():
            log.critical(f'Reference file {arg.ref} does not exists.')
            return success, arg
    arg.out = get_output(arg)
    if arg.out is None:
        return success, arg
    if not utils.accessible(arg.out, 'folder'):
        log.critical(f'Fail to create {arg.out}. Please contact the '
                     f'administrator.')
        return success, arg
    if arg.mt_mode:
        arg.simple_validate = True
        log.warning(f'Given genome length range is {arg.min}-{arg.max}.')
        log.warning('Is it ok for mitochondria?')
    arg.out.mkdir()
    arg.log = arg.out / 'Log'
    arg.log.mkdir()
    arg.raw = arg.out / 'Raw'
    arg.raw.mkdir()
    arg.tmp = arg.out / 'Temp'
    arg.tmp.mkdir()
    success, arg.third_party = utils.get_third_party()
    return success, arg


def read_table(list_file: Path):
    """
    Read table from given csv file.
    Columns of table:
        Input, Input(optional), Taxonomy
    Args:
        list_file(Path): csv file
    Return:
        inputs(list): [[f, r], taxon]
    """
    inputs = []
    with open(list_file, 'r') as raw:
        for line in raw:
            try:
                f, r, taxon = line.strip().split(',')
            except IndexError:
                log.warning(f'Cannot parse the line : {line}')
                continue
            if r == '':
                f_r = [f, ]
            else:
                f_r = [f, r]
            f_r = [Path(i).absolute() for i in f_r]
            inputs.append([f_r, taxon])
    return inputs


def split(raw, number, output: Path):
    """
    Split reads of original file from the beginning.
    If set number to default('inf'), extract all reads.
    If number is "inf" and format is not gz, skip split.
    Args:
        raw(str or Path): input file, coulde be fastq or gz format
        number(int or 'inf'): number of reads to split, 'inf' for no limit
        output(Path): output folder
    Return:
        splitted(Path): splitted file, fastq format
        count(int): reads actually got, 0 for not split
    """
    raw = Path(raw).absolute()
    fmt = utils.get_fmt(raw)
    if fmt != 'gz' and number == float('inf'):
        log.debug('Skip split for "inf" and non-gz.')
        return raw, 0
    splitted = output / raw.with_suffix(f'.{number}').name
    splitted_handle = open(splitted, 'wb')
    if fmt == 'gz':
        raw_handle = gzip.open(raw)
    else:
        raw_handle = open(raw, 'rb')
    line = iter(raw_handle)
    count = 0
    while count < number:
        # four line one record
        try:
            splitted_handle.write(next(line))
            splitted_handle.write(next(line))
            splitted_handle.write(next(line))
            splitted_handle.write(next(line))
        except StopIteration:
            break
        count += 1
    raw_handle.close()
    splitted_handle.close()
    splitted = utils.move(splitted, splitted.with_suffix(f'.{count}'))
    if number != float('inf') and number != count:
        log.warning(f'Want {number} reads, acutally got {count}.')
    return splitted, count


def get_reads_len(filename):
    """
    Get reads length of fastq
    """
    fmt = utils.get_fmt(filename)
    if fmt == 'gz':
        handle = gzip.open(filename)
    else:
        handle = open(filename, 'rb')
    handle.readline()
    seq = handle.readline()
    # normally it's safe to use utf8
    seq = seq.decode('utf-8').strip()
    length = len(seq)
    handle.close()
    log.debug(f'\tReads length: {length}')
    return length


def get_seed(ref: Path, output: Path, gene: str):
    """
    Use BarcodeFinder to get seed or reference sequence.
    Arg:
        ref(Path): reference chloroplast genome gb file, only contains one
        record
        output(Path): output folder
        gene(str): gene names, separated by comma
        mt_mode(bool): mt_mode or not, if, only use whole.seed
    Return:
        seeds(list): seed files list
    """
    ordered_seeds = []
    whole = output / 'whole.seed'
    SeqIO.convert(ref, 'gb', whole, 'fasta')
    seeds = {}
    genes = gene.split(',')
    gb = SeqIO.read(ref, 'gb')
    accession = gb.annotations['accessions'][0]
    organism = gb.annotations['organism'].replace(' ', '_')
    for feature in gb.features:
        if feature.type == 'gene' and 'gene' in feature.qualifiers:
            gene_name = feature.qualifiers['gene'][0]
            if gene_name in genes:
                seq = feature.extract(gb)
                seed_file = output / f'{gene_name}.seed'
                with open(seed_file, 'w') as out:
                    out.write(f'>{gene_name}|{organism}|{accession}\n')
                    out.write(f'{seq.seq}\n')
                seeds[gene_name] = seed_file
    for i in genes:
        if i in seeds:
            ordered_seeds.append(seeds[i])
    ordered_seeds.append(whole)
    return ordered_seeds


def config(seed: Path, arg):
    """
    Generate config file for NOVOPlasty.
    Arg:
        seed(Path): seed file
        arg(NameSpace): parameters user provided
    Return:
        config_file(Path): config file
    """
    if len(arg.input) == 2:
        f, r = arg.input
        m = ''
        s_or_p = 'PE'
    else:
        f = r = ''
        m = arg.input[0]
        s_or_p = 'SE'
    if arg.mt_mode:
        type_ = 'mito'
    else:
        type_ = 'chloro'
    arg.reads_len = get_reads_len(arg.input[0])
    if arg.insert_size is None:
        arg.insert_size = arg.reads_len * 2 + 50
        log.debug(f'The insert size is missing, use {arg.insert_size}.')
    config_str = f"""Project:
-----------------------
Project name          = {arg.out.name}
Type                  = {type_}
Genome Range          = {arg.min}-{arg.max}
K-mer                 = {arg.kmer}
Max memory            = {arg.mem}
Extended log          = 1
Extend seed directly  = no
Save assembled reads  = no
Seed Input            = {seed}
Reference sequence    =
Variance detection    = no
Chloroplast sequence  =

Dataset 1:
-----------------------
Read Length    = {arg.reads_len}
Insert size    = {arg.insert_size }
Platform       = {arg.platform}
Single/Paired  = {s_or_p}
Combined reads = {m}
Forward reads  = {f}
Reverse reads  = {r}

Optional:
-----------------------
Insert size auto      = yes
Insert Range          = 1.9
Insert Range strict   = 1.3
Use Quality Scores    = no
"""
    config_file = arg.raw / f'{seed.stem}_config.ini'
    with open(config_file, 'w') as out:
        out.write(config_str)
    return config_file


def organize_out(arg, seed: str):
    """
    Organize NOVOPlasty output.
        log*: log file
        contigs_tmp*: temporary files
        Contigs*: contigs
        Merged*: merged contigs, may be circular or empty, contains options
        Option*: merged contigs, circular or incomplete circular
        Circularized*: circularized sequence
    Return fasta list.
    Arg:
        arg(NameSpace): args
        seed(str): seed gene's name
    Return:
        contigs(list): contig files
        merged(list): merged files
        options(list): options files
        circularized(list): circularized files
    """
    def txt_to_fasta(old):
        """
        Convert NOVOPlasty-generated txt to standard fasta.
        """
        clean = []
        record = []
        begin = False
        with open(old, 'r') as raw:
            for line in raw:
                if line.startswith('>'):
                    clean.extend(record)
                    record = []
                    begin = True
                if line.startswith(' ') or len(line.strip()) == 0:
                    begin = False
                    clean.extend(record)
                    record = []
                if begin:
                    record.append(line)
        clean.extend(record)
        new = Path(old).with_suffix('.fasta')
        with open(new, 'w') as output:
            for line in clean:
                output.write(line.replace('*', ''))
        return new

    for i in arg.out.glob('contigs_tmp_*.txt'):
        i = i.absolute()
        utils.move(i, arg.tmp/i.with_name(f'{i.stem}-{seed}{i.suffix}').name)
    for i in arg.out.glob('log_*.txt'):
        i = i.absolute()
        utils.move(i, arg.log/i.with_name('NOVOPlasty-'+i.name).name)
    contigs = []
    for i in arg.out.glob('Contigs_*.fasta'):
        i = i.absolute()
        i = utils.move(i, arg.raw/i.with_name(f'{i.stem}-{seed}{i.suffix}').name)
        contigs.append(i)
    merged = []
    for i in arg.out.glob('Merged_contigs_*.txt'):
        i = i.absolute()
        i = utils.move(i, arg.raw/i.with_name(f'{i.stem}-{seed}{i.suffix}').name)
        fasta = txt_to_fasta(i)
        fasta = utils.move(fasta, arg.raw/fasta.name)
        merged.append(fasta)
    options = []
    for i in arg.out.glob('Option_*.fasta'):
        i = i.absolute()
        i = utils.move(i, arg.raw/i.with_name(f'{i.stem}-{seed}{i.suffix}').name)
        options.append(i)
    circularized = []
    for i in arg.out.glob('Circularized_assembly*.fasta'):
        i = i.absolute()
        i = utils.move(i, arg.raw/i.with_name(f'{i.stem}-{seed}{i.suffix}').name)
        circularized.append(i)
    return circularized, options, merged, contigs


def assembly(arg, perl, novoplasty):
    """
    Assembly input file by wrapping NOVOPlasty.
    The way to find absolute path of perl seems not good enough (I forget why
    but I remember I tried).
    Args:
        arg(NameSpace): arguments
        perl(str): perl location, may not be absolute path
        novoplasty(Path): novoplasty file
    Return:
        success(bool): success or not
    """
    def _patient_log():
        # hint user every 30s to avoid long time boring waiting
        n = 0
        while novoplasty_is_running:
            sleep(1)
            n += 1
            if n >= 30:
                log.info('NOVOPlasty is running, please be patient...')
                n = 0
        return

    success = False
    log.info('')
    for i in arg.input:
        test = Path(i)
        # arg.list_file may contains invalid file
        if not test.exists():
            log.critical(f'Cannot find input file {i}')
            return success
        else:
            log.info(f'Input file: {i}')
    log.info(f'Minimum genome size: {arg.min}')
    log.info(f'Maximum genome size: {arg.max}')
    if arg.ref is not None:
        log.info(f'Reference: {arg.ref}')
    elif isinstance(arg.taxon, list):
        t_ = ' '.join(arg.taxon)
        log.info(f'Taxonomy: {t_}')
    else:
        log.info(f'Taxonomy: {arg.taxon}')
    log.info(f'Output folder: {arg.out}')
    # split
    # equal to zero or not, expose to user
    # equal to inf or not, hide inside
    have_gz = ('gz' in [utils.get_fmt(i) for i in arg.input])
    if arg.split != 0:
        log.info(f'Split {arg.split} pairs of reads for assembly')
        splitted = []
        for raw in arg.input:
            new, count = split(raw, arg.split, arg.tmp)
            splitted.append(new)
        arg.input = splitted
    # novoplasty calls gzip, which Windows does not have
    elif platform.system() == 'Windows' and have_gz:
        log.debug(f'Split for gz on Windows.')
        splitted = []
        for raw in arg.input:
            new, count = split(raw, float('inf'), arg.tmp)
            splitted.append(new)
        arg.input = splitted
    # get ref
    if arg.ref is not None:
        if utils.get_fmt(arg.ref) != 'gb':
            log.critical(f'Reference file should be genbank format, '
                         f'but {arg.ref} is not.')
            return success
        ref = Path(arg.ref).absolute()
        ref = utils.move(ref, arg.tmp/ref.name, copy=True)
    else:
        # for "Genus species var. blabla", ignore subspecies words
        if isinstance(arg.taxon, str):
            pass
        elif isinstance(arg.taxon, list):
            arg.taxon = ' '.join(arg.taxon[:2])
        else:
            pass
        ref, arg.taxon = utils.get_ref(arg.taxon, arg.tmp, mt_mode=arg.mt_mode,
                                       simple_validate=arg.simple_validate)
        if ref is None:
            log.critical('Cannot get reference.')
            return success
        else:
            ref = utils.move(ref, arg.tmp/ref.name)
    # get seed
    seeds = []
    ordered_seeds = get_seed(ref, arg.raw, arg.seed)
    if arg.seed_file is not None:
        seeds.append(Path(arg.seed_file).absolute())
        # only add whole.seed
        seeds.append(ordered_seeds[-1])
    else:
        seeds.extend(ordered_seeds)
    if len(seeds) == 0:
        log.critical('Cannot get seeds!')
        return success
    csv_files = []
    all_contigs = []
    for seed in seeds:
        log.info(f'Use {seed.stem} as seed.')
        config_file = config(seed, arg)
        log.info('Call NOVOPlasty... May need minutes (rarely half an hour)')
        # use mark to terminate thread
        novoplasty_is_running = True
        hint = Thread(target=_patient_log)
        hint.start()
        # ignore bad returncode
        run(f'{perl} {novoplasty} -c {config_file}', shell=True,
            stdout=DEVNULL, stderr=DEVNULL)
        novoplasty_is_running = False

        # novoplasty use current folder as output folder
        circularized, options, merged, contigs = organize_out(arg, seed.stem)
        all_contigs.extend(contigs)
        if len(circularized) == 0 and len(options) == 0 and len(merged) == 0:
            log.warning(f'Assembled with {seed.stem} failed.')
            continue
        validated = []
        log.info('Validate assembly results.')
        # validate merged or not?
        for i in (*circularized, *options):
            arg_str = (f'-input {i} -ref {ref} -seed {seed.stem} '
                       f'-out {arg.out} '
                       f'-perc_identity {arg.perc_identity} '
                       f'-len_diff {arg.len_diff}')
            # if mt, simple validate
            if arg.simple_validate:
                arg_str += ' -simple_validate'
            validate_file, report = validate_main(arg_str)
            validated.extend(validate_file)
            if report not in csv_files:
                csv_files.append(report)
        if len(validated) != 0:
            success = True
            break
        else:
            log.warning('No records passed validation.')
        if not success:
            log.critical(f'Assembly with {seed.stem} failed.')
    if not success:
        log.info('Failed with all seeds.')
        all_input = ' '.join([str(i.absolute()) for i in all_contigs])
        # '' means empty
        if len(all_input) == 0:
            return success
        log.info('Try to assembly contigs generated from each seed.')
        arg_str = f'-input {all_input} -o {arg.out/"Raw"/"merge_seed.fasta"}'
        n_assembly, assembly_result = merge_main(arg_str)
        if n_assembly != 0:
            arg_str = (f'-input {assembly_result} -ref {ref} -seed merge '
                       f'-out {arg.out} '
                       f'-perc_identity {arg.perc_identity} '
                       f'-len_diff {arg.len_diff}')
            if arg.simple_validate:
                arg_str += ' -simple_validate'
            validate_file, report = validate_main(arg_str)
            if len(validate_file) != 0:
                success = True
                csv_files.append(report)
    return success


def assembly_main(arg_str=None):
    """
    Wrap function for assembly.
    Args:
        arg_str(str or None): string for parse_arg
    Return:
        success(bool): success or not
        arg.out(Path): output path, for UI only
    """
    log.info('Welcome to novowrap.')
    success = False
    # check arg
    if arg_str is None:
        arg = parse_args()
    else:
        arg = parse_args(arg_str.split(' '))
    success, arg = init_arg(arg)
    if not success:
        log.critical('Quit.')
        return success, arg.out
    else:
        log.debug('Init OK.')
    # check before run
    # seems cannot use thread to save time
    perl = utils.get_perl(arg.third_party)
    if perl == '':
        log.critical('Failed to get perl. Quit.')
        return success, arg.out
    novoplasty = utils.get_novoplasty(arg.third_party)
    if novoplasty is None:
        log.critical('Quit.')
        return success, arg.out
    # log to file
    log_file_handler = logging.FileHandler(str(arg.log/'Log.txt'))
    # more detail in file log
    # coloredlog and logging ahve differnt levels
    log.setLevel(logging.DEBUG)
    formatter = logging.Formatter(FMT, DATEFMT)
    log_file_handler.setFormatter(formatter)
    log.addHandler(log_file_handler)
    # start
    cwd = Path().cwd().absolute()
    if arg.list_file is None:
        # novoplasty put files in current folder, have to chdir to make
        # the folder clean
        chdir(arg.out)
        success = assembly(arg, perl, novoplasty)
    else:
        table = read_table(arg.list_file)
        success_list = []
        original_out = arg.out.absolute()
        for i in table:
            arg.input, arg.taxon = i
            # str to list
            if len(arg.taxon) == 0:
                arg.taxon = 'Nicotiana tabacum'
                log.warning(f'Taxonomy for {arg.input[0]} is missing.')
                log.info(f'Use {arg.taxon} instead.')
            arg.taxon = arg.taxon.split(' ')
            new_out = _get_name(arg.input)
            arg.out = original_out / new_out.stem
            arg.out.mkdir()
            chdir(arg.out)
            s = assembly(arg, perl, novoplasty)
            success_list.append(s)
            chdir(cwd)
        success = all(success_list)
    log.info('Bye.')
    chdir(cwd)
    return success, arg.out


if __name__ == '__main__':
    assembly_main()



#!/usr/bin/python3

from os import environ
from pathlib import Path
from random import randint
import argparse
import logging

from Bio import SeqIO
from matplotlib import use as mpl_use
# if environ.get('DISPLAY', '') == '':
mpl_use('Agg')
from matplotlib import pyplot as plt
import numpy as np

from novowrap import utils


# define logger
FMT = '%(asctime)s %(levelname)-8s %(message)s'
DATEFMT = '%H:%M:%S'
logging.basicConfig(format=FMT, datefmt=DATEFMT, level=logging.INFO)
try:
    import coloredlogs
    coloredlogs.install(level=logging.INFO, fmt=FMT, datefmt=DATEFMT)
except ImportError:
    pass
# inherit logger from novowrap, if not called by it, doesn't matter to
# name the logger 'novowrap'
log = logging.getLogger('novowrap')


def parse_args(arg_list=None):
    arg = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    arg.add_argument('-input', required=True, help='input filename')
    arg.add_argument('-ref', help='reference gb')
    arg.add_argument('-taxon', nargs='*', help='Taxonomy name')
    options = arg.add_argument_group('Option')
    options.add_argument('-simple_validate', action='store_true',
                         help='for plastids with abnormal structure')
    options.add_argument('-mt', dest='mt_mode', action='store_true',
                         help='for mitochondria (EXPERIMENTAL)')
    options.add_argument('-perc_identity', type=float, default=0.7,
                         help='minimum percentage of identity of BLAST, 0-100')
    options.add_argument('-len_diff', type=float, default=0.2,
                         help='maximum percentage of length difference of '
                         'query to reference, 0-100')
    options.add_argument('-seed',
                         help='seed used in assembly, only for caller')
    options.add_argument('-out', help='output folder')
    options.add_argument('-debug', action='store_true',
                         help='print debug info')
    if arg_list is None:
        return arg.parse_args()
    else:
        return arg.parse_args(arg_list)


def init_arg(arg):
    """
    Initialize working folder with arg.
    Args:
        arg(NameSpace): arg generated from parse_args
    Return:
        success(bool): success or not
        arg(NameSpace): arg with output info
    """
    if arg.debug:
        logging.basicConfig(level=logging.DEBUG)
        try:
            import coloredlogs
            coloredlogs.install(level=logging.DEBUG, fmt=FMT, datefmt=DATEFMT)
        except ImportError:
            pass
    success = False
    arg.input = Path(arg.input).absolute()
    if not arg.input.exists():
        log.critical(f'Input file {arg.input} does not exist.')
        return success, arg
    if arg.ref is None and arg.taxon is None:
        log.warning('Nor reference either taxonomy was given.')
        return success, arg
    elif arg.ref is not None and arg.taxon is not None:
        log.critical('Cannot use "-taxon" and "-ref" at same time.')
        return success, arg
    if arg.ref is not None:
        arg.ref = Path(arg.ref).absolute()
    if arg.taxon is None:
        pass
    elif len(arg.taxon) > 1:
        # for "Genus species var. blabla", ignore subspecies words
        arg.taxon = ' '.join(arg.taxon[:2])
    else:
        arg.taxon = arg.taxon[0]
    if arg.out is None:
        arg.out = Path(arg.input.stem+'-out').absolute()
    else:
        arg.out = Path(arg.out).absolute()
    if __name__ != '__main__':
        # if called from novowrap, out exist and accessible already
        pass
    else:
        if arg.out.exists():
            log.warning(f'Output folder {arg.out.name} exists.')
            # give users one more chance
            new_name = arg.out.name + f'-{randint(0, 2020):04d}'
            arg.out = arg.out.with_name(new_name)
            if arg.out.exists() or not utils.accessible(arg.out, 'folder'):
                log.critical('Cannot create output folder.')
                return success, arg
            else:
                arg.out.mkdir()
                log.info(f'Use {arg.out.name} instead.')
        else:
            if not utils.accessible(arg.out, 'folder'):
                log.critical(f'Failed to access output folder {arg.out}.'
                             f'Please contact the administrator.')
                return success, arg
            arg.out.mkdir()
    arg.tmp = arg.out / 'Temp'
    # if called by novowrap, tmp is accessible already
    if not arg.tmp.exists():
        arg.tmp.mkdir()
    if arg.mt_mode:
        arg.simple_validate = True
    success = True
    return success, arg


def process_ref(arg):
    """
    Preprocess reference
    Prefer ref because assembly passed ref
    Returns:
        r_ref_gb(Path): rotated gb
        r_ref_fasta(Path): rotated fasta
        ref_len(int): length of the reference
    """
    r_ref_gb = None
    r_ref_fasta = None
    ref_len = 0
    if arg.ref is not None:
        log.info(f'Reference:\t{arg.ref}')
        fmt = utils.get_fmt(arg.ref)
        ref_gb = Path(arg.ref).absolute()
        ref_gb = utils.move(ref_gb, arg.tmp/ref_gb.name, copy=True)
        ref_records = list(SeqIO.parse(ref_gb, fmt))
        if len(ref_records) > 1:
            log.warning('Given reference contains more than one records, '
                        'only use the first.')
            # assume given reference is ok since user refuse to auto download
            # reference from Genbank
            SeqIO.write(ref_records[0], ref_gb, fmt)
    else:
        log.info(f'Taxonomy:\t{arg.taxon}')
        ref_gb, ref_taxon = utils.get_ref(arg.taxon, arg.tmp)
        if ref_gb is None:
            log.critical('Failed to get reference.')
            log.debug(f'{arg.input} {arg.ref} REF_NOT_FOUND\n')
            return r_ref_gb, r_ref_fasta, ref_len
        ref_gb = utils.move(ref_gb, arg.tmp/ref_gb.name)
        fmt = 'gb'
    log.info(f'Output:\t {arg.out}')
    ref_len = len(SeqIO.read(ref_gb, fmt))
    r_ref_gb, r_ref_fasta = utils.rotate_seq(
        ref_gb, tmp=arg.tmp, simple_validate=arg.simple_validate)
    if r_ref_gb is None:
        log.critical('Cannot process reference sequence.')
        log.critical('Please consider to use another reference.')
        log.debug(f'{arg.input} {arg.ref} REF_CANNOT_ROTATE\n')
        return r_ref_gb, r_ref_fasta, ref_len
    return r_ref_gb, r_ref_fasta, ref_len


def divide_records(fasta: Path, output: Path, ref_len: int,
                   tmp: Path, len_diff=0.1, simple_validate=False):
    """
    Make sure each file has only one record.
    Args:
        fasta(Path): fasta file
        output(Path): output folder
        ref_len(int): length of reference, to filter bad records
        tmp(Path): temp folder
        len_diff: maximum allowed length difference
        simple_validate(bool): skip normal rotate or not
    Returns:
        divided(dict): info of divided files
    """
    options = list(SeqIO.parse(fasta, 'fasta'))
    divided = {}
    keys = ('gb,fasta,length,LSC,IRa,SSC,IRb,missing,incomplete,'
            'rc,figure,figure_after,skip').split(',')
    log.info(f'Found {len(options)} records in {fasta.name}.')
    if len(options) > 1:
        log.info('Divide them into different files.')
    log.info(f"Check record's length (reference: {ref_len} bp, "
             f"difference limit {len_diff:.2%}).")
    for idx, record in enumerate(options):
        skip = False
        if len(options) > 1:
            filename = output / f'{fasta.stem}_{idx+1}{fasta.suffix}'
        else:
            filename = output / fasta.name
        divided[filename] = dict((key, '') for key in keys)
        record_len = len(record)
        record_len_diff = (record_len/ref_len) - 1
        divided[filename]['fasta'] = filename
        divided[filename]['length'] = record_len
        divided[filename]['length_diff'] = record_len_diff
        if abs(record_len_diff) > len_diff:
            log.warning(f'Skip NO.{idx+1} record ({record_len} bp, '
                        f'length difference {record_len_diff:.2%}).')
            skip = 'undersize' if record_len_diff < 0 else 'oversize'
        SeqIO.write(record, filename, 'fasta')
        if not skip:
            r_gb, r_fasta = utils.rotate_seq(
                filename, tmp=tmp, simple_validate=simple_validate)
            if r_gb is not None:
                divided[filename].update({'gb': r_gb, 'fasta': r_fasta,
                                          'length': record_len})
                utils.move(filename, tmp/filename.with_suffix('.raw').name)
            else:
                skip = 'structure_unusual'
        divided[filename]['skip'] = skip
        log.debug(f'{skip}')
    return divided


def compare_seq(query, reference, tmp, perc_identity):
    """
    Use BLAST to compare two records.
    Args:
        query(Path or str): query file
        reference(Path or str): reference file
        tmp(Path): temp folder
        perc_identity(float): percent identity for BLAST, need multiply 100
    Return:
        results[0][1]: BLAST data
    """
    results = []
    blast_result, blast_log = utils.blast(Path(query), reference,
                                          perc_identity*100)
    if blast_result is None:
        return None
    # only one record in file, loop is for unpack
    for query in utils.parse_blast_tab(blast_result):
        record = []
        for i in query:
            (qseqid, sseqid, sstrand, qlen, slen, length, pident, gapopen,
             qstart, qend, sstart, send) = i
            record.append([qstart, qend, sstart, send, sstrand, pident])
        results.append(record)
    utils.move(blast_result, tmp/blast_result.name)
    utils.move(blast_log, tmp/blast_log.name)
    if len(results) == 0:
        return None
    else:
        return results[0]


def get_alpha(old):
    """
    Given 0-100, return 0, 0.5, 0.75, 0.95, 1
    Args:
        old(float): percent of identity
    Return:
        alpha(float): alpha value
    """
    if old < 50:
        alpha = 0
    elif old < 80:
        alpha = 0.1
    elif old < 95:
        alpha = 0.15
    elif old < 100:
        alpha = 0.2
    else:
        alpha = 0.3
    return alpha


def draw(ref_gb: Path, seq_gb: Path, data: list):
    """
    Draw figure.
    Args:
        ref_gb(Path): reference genbank file
        seq_gb(Path): sequence genbank file
        data(list): BLAST result
    Return:
        pdf(Path): figure file
    """
    ref_regions = utils.get_regions(ref_gb)
    seq_regions = utils.get_regions(seq_gb)
    title = f'{seq_gb.stem} and {ref_gb.stem}'
    ignore_offset = len(ref_regions['IRa'])*2 + len(ref_regions['SSC'])
    plt.rcParams.update({'font.size': 20, 'font.family': 'serif'})
    plt.figure(1, figsize=(30, 15))
    plt.title(f'Validation of {title}', pad=30)
    plt.xlabel('Base')
    y_max = max(ref_regions['IRb'].location.end,
                seq_regions['IRb'].location.end)
    for key, value in ref_regions.items():
        plt.plot([value.location.start, value.location.end], [0.8, 0.8],
                 marker='+', label=key, linewidth=10)
        plt.text(value.location.start+len(value)/2, 0.78, f'{len(value)} bp',
                 fontsize=20, ha='center')
    for key, value in seq_regions.items():
        plt.text(value.location.start+len(value)/2, 0.96, f'{len(value)} bp',
                 fontsize=20, ha='center')
        if key != 'IRb':
            plt.axvline(x=value.location.end, ymin=0.15, ymax=0.85,
                        linestyle='dashdot', alpha=0.5)
    # no repeat legend
    plt.plot(0.5, 0.5, 'r-+', linewidth=5, label='Plus')
    plt.plot(0.5, 0.5, 'g-|', linewidth=5, label='Minus')
    plt.ylim([0.5, 1.1])
    plt.xlim([0, y_max])
    # usually fine
    if y_max > 20000:
        plt.xticks(list(range(0, y_max, 20000)))
    plt.yticks([0.65, 0.8, 0.95], labels=['Minus', 'Reference', 'Plus'])
    plt.legend(loc='lower left')
    for i in data:
        qstart, qend, sstart, send, sstrand, pident = i
        if sstrand == 'plus':
            plt.plot([qstart, qend], [0.95, 0.95], 'r-+', linewidth=5)
            # ignore these line
            if abs(qstart-sstart) > ignore_offset:
                continue
            plt.fill([sstart, qstart, qend, send], [0.8, 0.95, 0.95, 0.8],
                     color='r', alpha=get_alpha(pident))
        # minus strand
        else:
            plt.plot([qstart, qend], [0.65, 0.65], 'g-|', linewidth=5)
            plt.fill([send, sstart, qend, qstart], [0.8, 0.8, 0.65, 0.65],
                     color='#88cc88', alpha=get_alpha(pident))
    # pdf = seq_gb.with_name(seq_gb.stem+'.pdf')
    pdf = seq_gb.with_suffix('.pdf')
    plt.savefig(pdf)
    plt.close()
    return pdf


def simple_draw(ref_gb: Path, seq_gb: Path, data: list):
    """
    Draw figure of sequences without 4-parts structure.
    Args:
        ref_gb(Path): reference genbank file
        seq_gb(Path): sequence genbank file
        data(list): BLAST result
    Return:
        pdf(Path): figure file
    """
    ref_len = len(SeqIO.read(ref_gb, 'gb'))
    seq_len = len(SeqIO.read(seq_gb, 'gb'))
    title = f'{seq_gb.stem} and {ref_gb.stem}'
    ignore_offset = min(ref_len, seq_len)
    plt.rcParams.update({'font.size': 20, 'font.family': 'serif'})
    plt.figure(1, figsize=(30, 15))
    plt.title(f'Validation of {title}', pad=30)
    plt.xlabel('Base')
    y_max = max(ref_len, seq_len)
    plt.plot([0, ref_len], [0.8, 0.8], marker='+', label='Reference',
             linewidth=10)
    plt.text(ref_len/2, 0.78, f'{ref_len} bp', fontsize=20, ha='center')
    plt.text(ref_len/2, 0.96, f'{seq_len} bp', fontsize=20, ha='center')
    # no repeat legend
    plt.plot(0.5, 0.5, 'r-+', linewidth=5, label='Plus')
    plt.plot(0.5, 0.5, 'g-|', linewidth=5, label='Minus')
    plt.ylim([0.5, 1.1])
    plt.xlim([0, y_max])
    # usually fine
    if y_max > 20000:
        plt.xticks(list(range(0, y_max, 20000)))
    plt.yticks([0.65, 0.8, 0.95], labels=['Minus', 'Reference', 'Plus'])
    plt.legend(loc='lower left')
    for i in data:
        qstart, qend, sstart, send, sstrand, pident = i
        if sstrand == 'plus':
            plt.plot([qstart, qend], [0.95, 0.95], 'r-+', linewidth=5)
            # ignore these line
            if abs(qstart-sstart) > ignore_offset:
                continue
            plt.fill([sstart, qstart, qend, send], [0.8, 0.95, 0.95, 0.8],
                     color='r', alpha=get_alpha(pident))
        # minus strand
        else:
            plt.plot([qstart, qend], [0.65, 0.65], 'g-|', linewidth=5)
            plt.fill([send, sstart, qend, qstart], [0.8, 0.8, 0.65, 0.65],
                     color='#88cc88', alpha=get_alpha(pident))
    # pdf = seq_gb.with_name(seq_gb.stem+'.pdf')
    pdf = seq_gb.with_suffix('.pdf')
    plt.savefig(pdf)
    plt.close()
    return pdf


def validate_regions(length: int, regions: dict, compare: list,
                     perc_identity=0.7):
    """
    Use BLAST results to validate regions.
    Args:
        length(int): length of whole sequences
        regions(dict): regions information
        compare(list): BLAST result
        perc_identity: threshold of region similarity
    Returns:
        count(dict): count information
        to_rc(str or None): regions need to reverse-complement
        strand_info(dict): region's strand information
    """
    count = {}
    bad_region = False
    plus = np.zeros(length, dtype=bool)
    minus = np.zeros(length, dtype=bool)
    for region in regions:
        count[region] = {'plus': 0, 'minus': 0}
    for hsp in compare:
        qstart, qend, sstart, send, sstrand, pident = hsp
        if sstrand == 'plus':
            plus[qstart-1:qend] = True
        else:
            minus[qstart-1:qend] = True
    # count bases
    for rgn in count:
        start = int(regions[rgn].location.start)
        end = int(regions[rgn].location.end)
        p_slice = plus[start:end]
        m_slice = minus[start:end]
        count[rgn]['plus'] = np.count_nonzero(p_slice)
        count[rgn]['minus'] = np.count_nonzero(m_slice)
        count[rgn]['union'] = np.count_nonzero(p_slice | m_slice)
    # assign strand
    for rgn in count:
        # also use arg.perc_identity for region
        min_rgn_len = len(regions[rgn]) * perc_identity
        p = count[rgn]['plus']
        m = count[rgn]['minus']
        u = count[rgn]['union']
        if p == m == 0:
            count[rgn]['strand'] = 'missing'
        elif u < min_rgn_len:
            count[rgn]['strand'] = 'incomplete'
        elif p >= min_rgn_len:
            count[rgn]['strand'] = 'plus'
        elif m >= min_rgn_len:
            count[rgn]['strand'] = 'minus'
        else:
            count[rgn]['strand'] = 'unknown'
    # do not rc IR
    to_rc = None
    if count['LSC']['strand'] == count['SSC']['strand'] == 'minus':
        to_rc = 'whole'
    elif count['LSC']['strand'] == 'minus':
        to_rc = 'LSC'
    elif count['SSC']['strand'] == 'minus':
        to_rc = 'SSC'
    strand_info = {}
    for rgn in count:
        if count[rgn]['strand'] == 'missing':
            if strand_info.get('missing', '') == '':
                strand_info['missing'] = rgn
            else:
                strand_info['missing'] += f'-{rgn}'
            log.critical(f'Region {rgn} is missing.')
            bad_region = True
        elif count[rgn]['strand'] == 'incomplete':
            if strand_info.get('incomplete', '') == '':
                strand_info['incomplete'] = rgn
            else:
                strand_info['incomplete'] += f'-{rgn}'
            log.critical(f'Region {rgn} is incomplete.')
            bad_region = True
        else:
            pass
    return count, to_rc, strand_info, bad_region


def write_output(divided, r_ref_fasta, ref_len, ref_regions, arg):
    """
    Write output info.
    Args:
        divided:
        r_ref_fasta:
        ref_len:
        ref_regions:
        arg:
    Returns:
        output_info(Path): csv file
    """
    output_info = arg.out / f'{arg.out.name}-results.csv'
    if not output_info.exists():
        with open(output_info, 'w') as csv_head:
            csv_head.write('Input,Success,Seed,Length,LSC,IRa,SSC,IRb,'
                           'Missing,Incomplete,RC_region,'
                           'Reference,Ref_length,r_LSC,r_IRa,r_SSC,r_IRb\n')
    with open(output_info, 'a') as out:
        for record in divided:
            # format is easier than f-string for dict
            simple = divided[record]
            # add seed info
            simple['seed'] = str(arg.seed)
            simple['fasta'] = simple['fasta'].stem
            out.write('{fasta},{success},{seed},{length},{LSC},'
                      '{IRa},{SSC},{IRb},{missing},{incomplete},'
                      '{rc},'.format(**simple))
            out.write('{},{},{},{},{},{}\n'.format(
                r_ref_fasta.stem, ref_len, len(ref_regions['LSC']),
                len(ref_regions['IRa']), len(ref_regions['SSC']),
                len(ref_regions['IRb'])))
    return output_info


def normal_validate(divided: dict, r_ref_gb: Path, r_ref_fasta: Path, arg):
    """
    For chloroplast genomes with 4-parts structure
    """
    for i in divided:
        success = False
        divided[i]['success'] = success
        if divided[i]['skip']:
            continue
        i_gb = divided[i]['gb']
        i_fasta = divided[i]['fasta']
        log.info(f'Analyze {i_fasta}.')
        option_regions = utils.get_regions(i_gb)
        # add regions info
        for _ in option_regions:
            divided[i][_] = len(option_regions[_])
        compare_result = compare_seq(i_fasta, r_ref_fasta, arg.tmp,
                                     arg.perc_identity)
        if compare_result is None:
            log.critical('Bad BLAST results.')
            log.debug(f'{arg.input} {arg.ref} BLAST_FAIL\n')
            return None
        pdf = draw(r_ref_gb, i_gb, compare_result)
        utils.move(pdf, arg.out/pdf.name)
        log.info('Detecting reverse complement region.')
        option_len = divided[i]['length']
        count, to_rc, strand_info, bad_region = validate_regions(
            option_len, option_regions, compare_result, arg.perc_identity)
        divided[i].update(strand_info)
        if bad_region:
            # skip sequences with bad region
            continue
        if to_rc is not None:
            log.warning(f'Reverse complement the {to_rc} of {i_fasta.name}.')
            rc_fasta = utils.rc_regions(i_gb, to_rc)
            # clean old files
            utils.move(i_fasta, arg.tmp/(i_fasta.with_name(
                i_fasta.stem+'-noRC.fasta')).name)
            utils.move(i_gb, arg.tmp/(i_gb.with_name(
                i_gb.stem+'-noRC.gb')).name)
            rc_fasta = utils.move(rc_fasta, rc_fasta.with_suffix(''))
            r_rc_gb, r_rc_fasta = utils.rotate_seq(
                rc_fasta, tmp=arg.tmp, simple_validate=arg.simple_validate)
            if r_rc_gb is None:
                continue
            rc_fasta.unlink()
            r_rc_gb = utils.move(r_rc_gb, arg.out/r_rc_gb.with_name(
                r_rc_gb.stem+'_RC.gb').name)
            r_rc_fasta = utils.move(r_rc_fasta, arg.out/r_rc_fasta.with_name(
                r_rc_fasta.stem+'_RC.fasta').name)
            new_compare_result = compare_seq(r_rc_fasta, r_ref_fasta, arg.tmp,
                                             arg.perc_identity)
            pdf = draw(r_ref_gb, r_rc_gb, new_compare_result)
            utils.move(pdf, arg.out/pdf.name)
            divided[i]['fasta'] = r_rc_fasta
            new_regions = utils.get_regions(r_rc_gb)
            for _ in new_regions:
                divided[i][_] = len(new_regions[_])
            # validate again
            count_2, to_rc_2, *_ = validate_regions(
                option_len, new_regions, new_compare_result, arg.perc_identity)
            if to_rc_2 is None:
                success = True
        else:
            utils.move(i_fasta, i_fasta.with_suffix('.fasta'))
            success = True
        divided[i]['success'] = success
        return divided


def simple_validate(divided: dict, r_ref_gb: Path, r_ref_fasta: Path, arg):
    """
    For chloroplast genomes WITHOUT 4-parts structure
    """
    for i in divided:
        success = False
        divided[i]['success'] = success
        if divided[i]['skip']:
            continue
        i_gb = divided[i]['gb']
        i_fasta = divided[i]['fasta']
        log.info(f'Analyze {i_fasta}.')
        option_regions = utils.get_regions(i_gb)
        # add regions info
        for _ in option_regions:
            divided[i][_] = len(option_regions[_])
        compare_result = compare_seq(i_fasta, r_ref_fasta, arg.tmp,
                                     arg.perc_identity)
        if compare_result is None:
            log.critical('Cannot run BLAST.')
            log.debug(f'{arg.input} {arg.ref} BLAST_FAIL\n')
            return None
        pdf = simple_draw(r_ref_gb, i_gb, compare_result)
        utils.move(pdf, arg.out/pdf.name)
        log.debug('Skip detecting reverse complement region.')
        utils.move(i_fasta, i_fasta.with_suffix('.fasta'))
        success = True
        divided[i]['success'] = success
        return divided


def validate_main(arg_str=None):
    """
    Use BLAST to validate assembly result.
    Args:
        arg_str(str): arguments string
    Return:
        validated(list): list contains validated rotated fasta files
        output_info(str): result csv, empty string for failed result
    """
    validated = []
    output_info = ''
    if arg_str is None:
        arg = parse_args()
    else:
        arg = parse_args(arg_str.split(' '))
    init_ok, arg = init_arg(arg)
    if not init_ok:
        log.critical('Quit.')
        return validated, output_info
    log.info(f'Input:\t{arg.input}')
    r_ref_gb, r_ref_fasta, ref_len = process_ref(arg)
    if r_ref_gb is None:
        return validated, output_info

    ref_regions = utils.get_regions(r_ref_gb)
    divided = divide_records(arg.input, arg.out, ref_len, arg.tmp,
                             arg.len_diff, arg.simple_validate)
    if arg.simple_validate:
        divided = simple_validate(divided, r_ref_gb, r_ref_fasta, arg)
    else:
        divided = normal_validate(divided, r_ref_gb, r_ref_fasta, arg)
    if divided is None:
        return validated, output_info
    for i in divided:
        if divided[i]['success']:
            v_file = divided[i]['fasta']
            validated.append(v_file)
    if len(validated) != 0:
        log.info('Validated sequences:')
        for i in validated:
            log.info(f'\t{i.name}')
    output_info = write_output(divided, r_ref_fasta, ref_len, ref_regions, arg)
    log.info(f'Validation result was written into {output_info}')
    return validated, output_info


if __name__ == '__main__':
    validate_main()
